import typing as ty
import re
from pathlib import Path
from copy import copy
from typing import Self
import attrs.validators
from pydra.utils.typing import is_optional, is_fileset_or_union, is_truthy_falsy
from pydra.utils.general import get_fields
from pydra.utils.typing import StateArray, is_lazy
from pydra.utils.hash import hash_function
import os
import itertools
from collections import Counter
import attrs
import cloudpickle as cp
from pydra.utils.messenger import AuditFlag, Messenger
from pydra.utils.general import (
    attrs_fields,
    attrs_values,
)
from pydra.utils.hash import Cache, hash_single, register_serializer
from .field import Field, Arg, Out


if ty.TYPE_CHECKING:
    from pydra.engine.job import Job
    from pydra.environments.base import Environment
    from pydra.workers.base import Worker
    from pydra.engine.result import Result
    from pydra.engine.hooks import TaskHooks

TaskType = ty.TypeVar("TaskType", bound="Task")


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class Outputs:
    """Base class for all output definitions"""

    RESERVED_FIELD_NAMES = ("inputs",)
    BASE_ATTRS = ()

    _cache_dir: Path = attrs.field(default=None, init=False, repr=False)
    _node = attrs.field(default=None, init=False, repr=False)

    @property
    def inputs(self):
        """The inputs object associated with a lazy-outputs object"""
        if self._node is None:
            raise AttributeError(
                f"{self} outputs object is not a lazy output of a workflow node"
            )
        return self._node.inputs

    @classmethod
    def _from_job(cls, job: "Job[TaskType]") -> Self:
        """Collect the outputs of a job. This is just an abstract base method that
        should be used by derived classes to set default values for the outputs.

        Parameters
        ----------
        job : Job[TaskType]
            The job whose outputs are being collected.

        Returns
        -------
        outputs : Outputs
            The outputs of the job
        """
        defaults = {}
        for output in get_fields(cls):
            if output.mandatory:
                default = attrs.NOTHING
            elif isinstance(output.default, attrs.Factory):
                default = output.default.factory()
            else:
                default = output.default
            defaults[output.name] = default
        outputs = cls(**defaults)
        outputs._cache_dir = job.cache_dir
        return outputs

    @property
    def _results(self) -> "Result[Self]":
        results_path = self._cache_dir / "_job.pklz"
        if not results_path.exists():
            raise FileNotFoundError(f"Job results file {results_path} not found")
        with open(results_path, "rb") as f:
            return cp.load(f)

    def __iter__(self) -> ty.Generator[str, None, None]:
        """The names of the fields in the output object"""
        return iter(sorted(f.name for f in attrs_fields(self)))

    def __getitem__(self, name_or_index: str | int) -> ty.Any:
        """Return the value for the given attribute

        Parameters
        ----------
        name : str
            the name of the attribute to return

        Returns
        -------
        Any
            the value of the attribute
        """
        if isinstance(name_or_index, int):
            return list(self)[name_or_index]
        try:
            return getattr(self, name_or_index)
        except AttributeError:
            raise KeyError(
                f"{self} doesn't have an attribute {name_or_index}"
            ) from None

    def __eq__(self, other: ty.Any) -> bool:
        """Check if two tasks are equal"""
        values = attrs.asdict(self)
        fields = get_fields(self)
        try:
            other_values = attrs.asdict(other)
        except AttributeError:
            return False
        try:
            other_fields = get_fields(other)
        except AttributeError:
            return False
        if fields != other_fields:
            return False
        for field in get_fields(self):
            if field.hash_eq:
                values[field.name] = hash_function(values[field.name])
                other_values[field.name] = hash_function(other_values[field.name])
        return values == other_values

    def __repr__(self) -> str:
        """A string representation of the task"""
        fields_str = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}"
            for f in get_fields(self)
            if getattr(self, f.name) != f.default
        )
        return f"{self.__class__.__name__}({fields_str})"


OutputsType = ty.TypeVar("OutputType", bound=Outputs)


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class Task(ty.Generic[OutputsType]):
    """Base class for all tasks"""

    # Task type to be overridden in derived classes
    @classmethod
    def _task_type(cls) -> str:
        for base in cls.__mro__:
            parts = base.__module__.split(".")
            if parts[:2] == ["pydra", "compose"]:
                return parts[2]
        raise RuntimeError(
            f"Cannot determine task type for {cls.__name__} in module {cls.__module__} "
            "because none of its base classes are in the pydra.compose namespace:\n"
            + "\n".join(f"{b.__name__!r} in {b.__module__!r}" for b in cls.__mro__)
        )

    # The attribute containing the function/executable used to run the task
    _executor_name = None

    # Class attributes
    TASK_CLASS_ATTRS = ("xor",)
    _xor: frozenset[frozenset[str | None]] = (
        frozenset()
    )  # overwritten in derived classes

    # The following fields are used to store split/combine state information
    _splitter = attrs.field(default=None, init=False, repr=False)
    _combiner = attrs.field(default=None, init=False, repr=False)
    _container_ndim = attrs.field(default=None, init=False, repr=False)
    _hashes = attrs.field(default=None, init=False, eq=False, repr=False)

    RESERVED_FIELD_NAMES = ("split", "combine")
    BASE_ATTRS = ()

    def __call__(
        self,
        /,
        cache_root: os.PathLike | None = None,
        worker: "str | ty.Type[Worker] | Worker" = "debug",
        environment: "Environment | None" = None,
        rerun: bool = False,
        readonly_caches: ty.Iterable[os.PathLike] | None = None,
        audit_flags: AuditFlag = AuditFlag.NONE,
        messengers: ty.Iterable[Messenger] | None = None,
        messenger_args: dict[str, ty.Any] | None = None,
        hooks: "TaskHooks | None" = None,
        **kwargs: ty.Any,
    ) -> OutputsType:
        """Create a job from this task and execute it to produce a result.

        Parameters
        ----------
        cache_root : os.PathLike, optional
            Cache directory where the working directory/results for the job will be
            stored, by default None
        worker : str or Worker, optional
            The worker to use, by default "cf"
        environment: Environment, optional
            The execution environment to use, by default None
        rerun : bool, optional
            Whether to force the re-computation of the job results even if existing
            results are found, by default False
        readonly_caches : list[os.PathLike], optional
            Alternate cache locations to check for pre-computed results, by default None
        audit_flags : AuditFlag, optional
            Configure provenance tracking. available flags: :class:`~pydra.utils.messenger.AuditFlag`
            Default is no provenance tracking.
        messenger : :class:`Messenger` or :obj:`list` of :class:`Messenger` or None
            Messenger(s) used by Audit. Saved in the `audit` attribute.
            See available flags at :class:`~pydra.utils.messenger.Messenger`.
        messengers_args : messengers_args : dict[str, Any], optional
            Argument(s) used by `messegner`. Saved in the `audit` attribu
        **kwargs : dict
            Keyword arguments to pass on to the worker initialisation

        Returns
        -------
        OutputsType or list[OutputsType]
            The output interface of the job, or in the case of split tasks, a list of
            output interfaces
        """
        from pydra.engine.submitter import (  # noqa: F811
            Submitter,
            WORKER_KWARG_FAIL_NOTE,
        )

        try:
            with Submitter(
                audit_flags=audit_flags,
                cache_root=cache_root,
                readonly_caches=readonly_caches,
                messenger_args=messenger_args,
                messengers=messengers,
                environment=environment,
                worker=worker,
                **kwargs,
            ) as sub:
                result = sub(
                    self,
                    hooks=hooks,
                    rerun=rerun,
                )
        except TypeError as e:
            # Catch any inadvertent passing of task parameters to the
            # execution call
            if hasattr(e, "__notes__") and WORKER_KWARG_FAIL_NOTE in e.__notes__:
                if match := re.match(
                    r".*got an unexpected keyword argument '(\w+)'", str(e)
                ):
                    if match.group(1) in self:
                        e.add_note(
                            f"Note that the unrecognised argument, {match.group(1)!r}, is "
                            f"an input of the task {self!r} that has already been "
                            f"parameterised (it is being called to execute it)"
                        )
            raise
        if result.errored:
            if result.errors:
                time_of_crash = result.errors["time of crash"]
                error_message = "\n".join(result.errors["error message"])
            else:
                time_of_crash = "UNKNOWN-TIME"
                error_message = "NOT RETRIEVED"
            raise RuntimeError(
                f"Job {self} failed @ {time_of_crash} with the "
                f"following errors:\n{error_message}\n"
                "To inspect, please load the pickled job object from here: "
                f"{result.cache_dir}/_job.pklz"
            )
        return result.outputs

    def split(
        self,
        splitter: ty.Union[str, ty.List[str], ty.Tuple[str, ...], None] = None,
        /,
        overwrite: bool = False,
        container_ndim: ty.Optional[dict] = None,
        **inputs,
    ) -> Self:
        """
        Run this job parametrically over lists of split inputs.

        Parameters
        ----------
        splitter : str or list[str] or tuple[str] or None
            the fields which to split over. If splitting over multiple fields, lists of
            fields are interpreted as outer-products and tuples inner-products. If None,
            then the fields to split are taken from the keyword-arg names.
        overwrite : bool, optional
            whether to overwrite an existing split on the node, by default False
        container_ndim : dict, optional
            Container dimensions for specific inputs, used in the splitter.
            If input name is not in container_ndim, it is assumed that the input values has
            a container dimension of 1, so only the most outer dim will be used for splitting.
        **inputs
            fields to split over, will be automatically wrapped in a StateArray object
            and passed to the node inputs

        Returns
        -------
        self : TaskBase
            a reference to the job
        """
        from pydra.engine.state import unwrap_splitter
        from pydra.engine import lazy

        if self._splitter and not overwrite:
            raise ValueError(
                f"Cannot overwrite existing splitter {self._splitter} on {self}, "
                "set 'overwrite=True' to do so"
            )
        if splitter:
            unwraped_split = list(unwrap_splitter(splitter))
            if duplicated := [f for f, c in Counter(unwraped_split).items() if c > 1]:
                raise ValueError(f"Splitter fields {duplicated} are duplicated")
            split_names = set(
                s for s in unwraped_split if not s.startswith("_") and "." not in s
            )
            input_names = set(inputs)
            if missing_inputs := list(split_names - input_names):
                raise ValueError(
                    f"Splitter fields {missing_inputs} need to be provided as a keyword "
                    f"arguments to the split method (provided {list(inputs)})"
                )
            if unrecognised_inputs := list(input_names - split_names):
                raise ValueError(
                    f"Provided inputs {unrecognised_inputs} are not present in the "
                    f"splitter {splitter}"
                )
        else:
            # If no splitter is provided, use the names of the inputs as combinatorial splitter
            split_names = splitter = list(inputs)
        for field_name in container_ndim or []:
            if field_name not in split_names:
                raise ValueError(
                    f"Container dimension for {field_name} is provided but the field "
                    f"is not present in the inputs"
                )
        split_inputs = {}
        for name, value in inputs.items():
            if isinstance(value, lazy.LazyField):
                split_val = value.split()
            elif isinstance(value, ty.Iterable) and not isinstance(
                value, (ty.Mapping, str)
            ):
                split_val = StateArray(value)
            else:
                raise TypeError(
                    f"Could not split {value!r} as it is not a sequence type"
                )
            split_inputs[name] = split_val
        split_def = attrs.evolve(self, **split_inputs)
        split_def._splitter = splitter
        split_def._container_ndim = container_ndim
        return split_def

    def combine(
        self,
        combiner: ty.Union[ty.List[str], str],
        overwrite: bool = False,
    ) -> Self:
        """
        Combine inputs parameterized by one or more previous tasks.

        Parameters
        ----------
        combiner : list[str] or str
            the field or list of inputs to be combined (i.e. not left split) after the
            job has been run
        overwrite : bool
            whether to overwrite an existing combiner on the node
        **kwargs : dict[str, Any]
            values for the job that will be "combined" before they are provided to the
            node

        Returns
        -------
        self : Self
            a reference to the outputs object
        """
        if self._combiner and not overwrite:
            raise ValueError(
                f"Attempting to overwrite existing combiner {self._combiner} on {self}, "
                "set 'overwrite=True' to do so"
            )
        if isinstance(combiner, str):
            combiner = [combiner]
        local_names = set(c for c in combiner if "." not in c and not c.startswith("_"))
        if unrecognised := local_names - set(self):
            raise ValueError(
                f"Combiner fields {unrecognised} are not present in the task"
            )
        combined_def = copy(self)
        combined_def._combiner = combiner
        return combined_def

    def __repr__(self) -> str:
        """A string representation of the task"""
        fields_str = ", ".join(
            f"{f.name}={getattr(self, f.name)!r}"
            for f in get_fields(self)
            if getattr(self, f.name) != f.default
        )
        return f"{self.__class__.__name__}({fields_str})"

    def __iter__(self) -> ty.Generator[str, None, None]:
        """Iterate through all the names in the task"""
        return (
            f.name
            for f in get_fields(self)
            if not (f.name.startswith("_") or f.name in self.RESERVED_FIELD_NAMES)
        )

    def __eq__(self, other: ty.Any) -> bool:
        """Check if two tasks are equal"""
        values = attrs.asdict(self, recurse=False)
        if not isinstance(other, Task):
            return False
        try:
            other_values = attrs.asdict(other, recurse=False)
        except AttributeError:
            return False
        if set(values) != set(other_values):
            return False  # Return if attribute keys don't match
        for field in get_fields(self):
            if field.hash_eq:
                values[field.name] = hash_function(values[field.name])
                other_values[field.name] = hash_function(other_values[field.name])
        if values != other_values:
            return False
        hash_cache = Cache()
        if hash_function(type(self), cache=hash_cache) != hash_function(
            type(other), cache=hash_cache
        ):
            return False
        try:
            other_outputs = other.Outputs
        except AttributeError:
            return False
        return hash_function(self.Outputs, cache=hash_cache) == hash_function(
            other_outputs, cache=hash_cache
        )

    def __getitem__(self, name: str) -> ty.Any:
        """Return the value for the given attribute, resolving any templates

        Parameters
        ----------
        name : str
            the name of the attribute to return

        Returns
        -------
        Any
            the value of the attribute
        """
        try:
            return getattr(self, name)
        except AttributeError:
            raise KeyError(f"{self} doesn't have an attribute {name}") from None

    @property
    def _hash(self):
        hsh, self._hashes = self._compute_hashes()
        return hsh

    @property
    def _checksum(self):
        return f"{self._task_type()}-{self._hash}"

    def _hash_changes(self):
        """Detects any changes in the hashed values between the current inputs and the
        previously calculated values"""
        _, new_hashes = self._compute_hashes()
        return [k for k, v in new_hashes.items() if v != self._hashes[k]]

    def _compute_hashes(self) -> ty.Tuple[bytes, ty.Dict[str, bytes]]:
        """Compute a basic hash for any given set of fields."""
        inp_dict = {}
        for field in get_fields(self):
            if isinstance(field, Out):
                continue  # Skip output fields
            # removing values that are not set from hash calculation
            if getattr(self, field.name) is attrs.NOTHING:
                continue
            if getattr(field, "container_path", False):
                continue
            inp_dict[field.name] = getattr(self, field.name)
        # Include the outputs class, just in case any names or types have changed
        inp_dict["Outputs"] = self.Outputs
        hash_cache = Cache()
        field_hashes = {
            k: hash_function(v, cache=hash_cache) for k, v in inp_dict.items()
        }
        return hash_function(sorted(field_hashes.items())), field_hashes

    def _rule_violations(self) -> list[str]:
        """Check rules and returns a list of errors."""

        field: Arg
        errors = []
        for field in get_fields(self):
            value = self[field.name]

            if is_lazy(value):
                continue

            if (
                value is attrs.NOTHING
                and not getattr(field, "path_template", False)
                and not field.readonly
            ):
                errors.append(f"Mandatory field {field.name!r} is not set")

            # Raise error if any required field is unset.
            if (
                not (
                    value is None
                    or value is False
                    or (
                        is_optional(field.type)
                        and is_fileset_or_union(field.type)
                        and value is True
                    )
                )
                and field.requires
                and not any(rs.satisfied(self) for rs in field.requires)
            ):
                if len(field.requires) > 1:
                    qualification = (
                        " at least one of the following requirements to be satisfied: "
                    )
                else:
                    qualification = ""
                errors.append(
                    f"{field.name!r} requires{qualification} {[str(r) for r in field.requires]}"
                )
        # Collect alternative fields associated with this field.
        for xor_set in self._xor:
            mutually_exclusive = {name: self[name] for name in xor_set if name}
            are_set = [f"{n}={v!r}" for n, v in mutually_exclusive.items() if v]
            if len(are_set) > 1:
                errors.append(
                    f"Mutually exclusive fields ({', '.join(sorted(are_set))}) are set "
                    "together"
                )
            elif not are_set and None not in xor_set:
                errors.append(
                    "At least one of the mutually exclusive fields should be set: "
                    + ", ".join(f"{n}={v!r}" for n, v in mutually_exclusive.items())
                )
        return errors

    def _check_rules(self):
        """Check if all rules are satisfied."""

        attrs.validate(self)

        if errors := self._rule_violations():
            raise ValueError(
                f"Found the following errors in job {self} task:\n" + "\n".join(errors)
            )

    @classmethod
    def _check_arg_refs(
        cls,
        inputs: list[Arg],
        outputs: list[Out],
        xor: frozenset[frozenset[str | None]],
    ) -> None:
        """
        Checks if all fields referenced in requirements and xor are present in the inputs
        are valid field names
        """
        field: Field
        input_names = set(inputs)
        for field in itertools.chain(inputs.values(), outputs.values()):
            if unrecognised := (
                set([r.name for s in field.requires for r in s]) - input_names
            ):
                raise ValueError(
                    "'Unrecognised' field names in referenced in the requirements "
                    f"of {field} " + str(list(unrecognised))
                )

        for xor_set in xor:
            if unrecognised := xor_set - (input_names | {None}):
                raise ValueError(
                    f"Unrecognised field names in referenced in the xor {xor_set}: "
                    + str(list(unrecognised))
                )
            for field_name in xor_set:
                if field_name is None:  # i.e. none of the fields being set is valid
                    continue
                type_ = inputs[field_name].type
                if not is_truthy_falsy(type_):
                    raise ValueError(
                        f"Fields included in a 'xor' ({field_name!r}) must be an optional type or a "
                        f"truthy/falsy type, not type {type_}"
                    )

    def _check_resolved(self):
        """Checks that all the fields in the task have been resolved"""
        if lazy_values := [n for n, v in attrs_values(self).items() if is_lazy(v)]:
            raise ValueError(
                f"Cannot execute {self} because the following fields "
                f"still have lazy values {lazy_values}"
            )


# def set_none_default_if_optional(field: Field) -> None:
#     if is_optional(field.type) and field.mandatory:
#         field.default = None


@register_serializer
def bytes_repr_task(obj: Task, cache: Cache) -> ty.Iterator[bytes]:
    yield f"task[{obj._task_type()}]:(".encode()
    for field in get_fields(obj):
        yield f"{field.name}=".encode()
        yield hash_single(getattr(obj, field.name), cache)
        yield b","
    yield b"_splitter="
    yield hash_single(obj._splitter, cache)
    yield b",_combiner="
    yield hash_single(obj._combiner, cache)
    yield b",_container_ndim="
    yield hash_single(obj._container_ndim, cache)
    yield b",_xor="
    yield hash_single(obj._xor, cache)
    yield b")"



"""Execution workers."""

import asyncio
import attrs
import sys
import abc
import inspect
import asyncio.subprocess as asp
import typing as ty
import logging
from pydra.engine.job import Job
from pydra.utils.general import get_plugin_classes
import pydra.workers

logger = logging.getLogger("pydra.worker")

if ty.TYPE_CHECKING:
    from pydra.engine.result import Result
    from pydra.compose import base

TaskType = ty.TypeVar("TaskType", bound="base.Task")


@attrs.define
class Worker(metaclass=abc.ABCMeta):
    """A base class for execution of tasks."""

    loop: asyncio.AbstractEventLoop = None

    def __getstate__(self) -> dict[str, ty.Any]:
        """Return state for pickling."""
        state = attrs.asdict(self, recurse=False)
        state["loop"] = None
        return state

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        for key, value in state.items():
            setattr(self, key, value)
        self.loop = None
        # Loop will be restored by submitter __setstate__

    @abc.abstractmethod
    def run(self, job: "Job[TaskType]", rerun: bool = False) -> "Result":
        """Return coroutine for job execution."""
        pass

    async def submit(self, job: "Job[TaskType]", rerun: bool = False) -> "Result":
        assert self.is_async, "Worker is not asynchronous, job should just be `run()`"
        if job.is_async:  # only for workflows at this stage and the foreseeable
            # These jobs are run in the primary process but potentially farm out
            # workflow jobs to other processes/job-schedulers
            return await job.run_async(rerun=rerun)
        else:
            return await self.run(job=job, rerun=rerun)

    def close(self):
        """Close this worker."""

    @property
    def is_async(self) -> bool:
        """Return whether the worker is asynchronous."""
        return inspect.iscoroutinefunction(self.run)

    @classmethod
    def available_plugins(cls) -> ty.Dict[str, ty.Type["Worker"]]:
        """Return all installed worker types"""
        return get_plugin_classes(pydra.workers, "Worker")

    @classmethod
    def plugin(cls, plugin_name: str) -> ty.Type["Worker"]:
        """Return a worker class by name."""
        try:
            return cls.available_plugins()[plugin_name.replace("-", "_")]
        except KeyError:
            raise ValueError(
                f"No worker matches {plugin_name!r}, check if there is a "
                f"plugin package called 'pydra-workers-{plugin_name}' that needs to be "
                "installed."
            )

    @classmethod
    def plugin_name(cls) -> str:
        """Return the name of the plugin."""
        try:
            plugin_name = cls._plugin_name
        except AttributeError:
            parts = cls.__module__.split(".")
            if parts[:-1] != ["pydra", "workers"]:
                raise ValueError(
                    f"Cannot infer plugin name of Worker ({cls}) from module path, as it "
                    f"isn't installed within `pydra.workers` ({cls.__module__}). "
                    "Please set the `_plugin_name` attribute on the class explicitly."
                )
            plugin_name = parts[-1].replace("_", "-")
        return plugin_name


async def read_and_display_async(*cmd, hide_display=False, strip=False):
    """
    Capture standard input and output of a process, displaying them as they arrive.

    Works line-by-line.

    """
    # start process
    process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asp.PIPE, stderr=asp.PIPE
    )

    stdout_display = sys.stdout.buffer.write if not hide_display else None
    stderr_display = sys.stderr.buffer.write if not hide_display else None
    # read child's stdout/stderr concurrently (capture and display)
    try:
        stdout, stderr = await asyncio.gather(
            read_stream_and_display(process.stdout, stdout_display),
            read_stream_and_display(process.stderr, stderr_display),
        )
    except Exception:
        process.kill()
        raise
    finally:
        # wait for the process to exit
        rc = await process.wait()
    if strip:
        return rc, stdout.strip(), stderr
    else:
        return rc, stdout, stderr


async def read_stream_and_display(stream, display):
    """
    Read from stream line by line until EOF, display, and capture the lines.

    See Also
    --------
    This `discussion on StackOverflow
    <https://stackoverflow.com/questions/17190221>`__.

    """
    output = []
    while True:
        line = await stream.readline()
        if not line:
            break
        output.append(line)
        if display is not None:
            display(line)  # assume it doesn't block
    return b"".join(output).decode()


def ensure_non_negative(value: int) -> int:
    if not value or value < 0:
        return 0
    return value



from .field import Field, Arg, Out, NO_DEFAULT
from .helpers import (
    ensure_field_objects,
    parse_doc_string,
    extract_function_inputs_and_outputs,
    check_explicit_fields_are_none,
    extract_fields_from_class,
    is_set,
    sanitize_xor,
)
from .task import Task, Outputs
from .builder import build_task_class

__all__ = [
    "Field",
    "Arg",
    "Out",
    "NO_DEFAULT",
    "ensure_field_objects",
    "parse_doc_string",
    "extract_function_inputs_and_outputs",
    "check_explicit_fields_are_none",
    "extract_fields_from_class",
    "is_set",
    "build_task_class",
    "Task",
    "Outputs",
]



import typing as ty
import inspect
import attrs
import re
from copy import copy
from pydra.utils.typing import is_type, is_optional
from pydra.utils.general import get_fields
from .field import Field, Arg, Out, NO_DEFAULT


if ty.TYPE_CHECKING:
    from .task import Task, Outputs


def is_set(value: ty.Any) -> bool:
    """Check if a value has been set."""
    return value not in (attrs.NOTHING, NO_DEFAULT)


def ensure_field_objects(
    arg_type: type[Arg],
    out_type: type[Out],
    doc_string: str | None = None,
    inputs: dict[str, Arg | type] | None = None,
    outputs: dict[str, Out | type] | None = None,
    input_helps: dict[str, str] | None = None,
    output_helps: dict[str, str] | None = None,
) -> tuple[dict[str, Arg], dict[str, Out]]:
    """Converts dicts containing input/output types into input/output, including any
    help strings to the appropriate inputs and outputs

    Parameters
    ----------
    arg_type : type
        The type of the input fields
    out_type : type
        The type of the output fields
    doc_string : str, optional
        The docstring of the function or class
    inputs : dict[str, Arg | type], optional
        The inputs to the function or class
    outputs : dict[str, Out | type], optional
        The outputs of the function or class
    input_helps : dict[str, str], optional
        The help strings for the inputs
    output_helps : dict[str, str], optional
        The help strings for the outputs

    Returns
    -------
    inputs : dict[str, Arg]
        The input fields with help strings added
    outputs : dict[str, Out]
        The output fields with help strings added
    """

    for input_name, arg in list(inputs.items()):
        if isinstance(arg, Arg):
            if arg.name is None:
                arg.name = input_name
            elif arg.name != input_name:
                raise ValueError(
                    "Name of the argument must be the same as the key in the "
                    f"dictionary. The argument name is {arg.name} and the key "
                    f"is {input_name}"
                )
            else:
                arg.name = input_name
            if not arg.help:
                arg.help = input_helps.get(input_name, "") if input_helps else ""
        elif is_type(arg):
            inputs[input_name] = arg_type(
                type=arg,
                name=input_name,
                help=input_helps.get(input_name, ""),
            )
        elif isinstance(arg, dict):
            arg_kwds = copy(arg)
            if "help" not in arg_kwds:
                arg_kwds["help"] = input_helps.get(input_name, "")
            inputs[input_name] = arg_type(
                name=input_name,
                **arg_kwds,
            )
        else:
            raise ValueError(
                f"Input {input_name} must be an instance of {Arg}, a type, or a dictionary "
                f" of keyword arguments to pass to {Arg}, not {arg}"
            )

    for output_name, out in list(outputs.items()):
        if isinstance(out, Out):
            if out.name is None:
                out.name = output_name
            elif out.name != output_name:
                raise ValueError(
                    "Name of the argument must be the same as the key in the "
                    f"dictionary. The argument name is {out.name} and the key "
                    f"is {output_name}"
                )
            else:
                out.name = output_name
            if not out.help:
                out.help = output_helps.get(output_name, "") if output_helps else ""
        elif is_type(out):
            outputs[output_name] = out_type(
                type=out,
                name=output_name,
                help=output_helps.get(output_name, ""),
            )
            if is_optional(out):
                outputs[output_name].default = None
        elif isinstance(out, dict):
            out_kwds = copy(out)
            if "help" not in out_kwds:
                out_kwds["help"] = output_helps.get(output_name, "")
            if "path_template" in out_kwds:
                from pydra.compose.shell import outarg

                out_type_ = outarg
            else:
                out_type_ = out_type
            outputs[output_name] = out_type_(
                name=output_name,
                **out_kwds,
            )
        elif isinstance(out, ty.Callable) and hasattr(out_type, "callable"):
            outputs[output_name] = out_type(
                name=output_name,
                type=ty.get_type_hints(out).get("return", ty.Any),
                callable=out,
                help=re.split(r"\n\s*\n", out.__doc__)[0] if out.__doc__ else "",
            )
        else:
            raise ValueError(
                f"Unrecognised value provided to outputs ({arg}), can be either {out_type} "
                "type" + (" or callable" if hasattr(out_type, "callable") else "")
            )

    return inputs, outputs


def extract_function_inputs_and_outputs(
    function: ty.Callable,
    arg_type: type[Arg],
    inputs: list[str | Arg] | dict[str, Arg | type] | None = None,
    outputs: list[str | Out] | dict[str, Out | type] | type | None = None,
) -> tuple[dict[str, type | Arg], dict[str, type | Out]]:
    """Extract input output types and output names from the function source if they
    aren't explicitly

    Parameters
    ----------
    function : callable
        The function to extract the inputs and outputs from
    arg_type : type
        The type of the input fields
    out_type : type
        The type of the output fields
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The inputs to the function
    outputs : list[str | Out] | dict[str, Out | type] | type | None
        The outputs of the function

    Returns
    -------
    inputs : dict[str, Arg]
        The input fields extracted from the function
    outputs : dict[str, Out]
        The output fields extracted from the function
    """
    # if undefined_symbols := get_undefined_symbols(
    #     function, exclude_signature_type_hints=True, ignore_decorator=True
    # ):
    #     raise ValueError(
    #         f"The following symbols are not defined within the scope of the function "
    #         f"{function!r}, {undefined_symbols}. Ensure that all imports are "
    #         "defined within the function scope so it is portable"
    #     )
    sig = inspect.signature(function)
    type_hints = ty.get_type_hints(function)
    input_types = {}
    input_defaults = {}
    has_varargs = False
    for p in sig.parameters.values():
        if p.kind is p.VAR_POSITIONAL or p.kind is p.VAR_KEYWORD:
            has_varargs = True
            continue
        input_types[p.name] = type_hints.get(p.name, ty.Any)
        if p.default is not inspect.Parameter.empty:
            input_defaults[p.name] = p.default
    if inputs is not None:
        if not isinstance(inputs, dict):
            if non_named_args := [
                i for i in inputs if not isinstance(i, Arg) or i.name is None
            ]:
                raise ValueError(
                    "Only named Arg objects should be provided as inputs (i.e. not names or "
                    "other objects should not be provided when wrapping/decorating a "
                    f"function: found {non_named_args} when wrapping/decorating {function!r}"
                )
            inputs = {i.name: i for i in inputs}
        if not has_varargs:
            if unrecognised := set(inputs) - set(input_types):
                raise ValueError(
                    f"Unrecognised input names ({unrecognised}) not present in the signature "
                    f"of the function {function!r}"
                )
        for inpt_name, type_ in input_types.items():
            try:
                inpt = inputs[inpt_name]
            except KeyError:
                inputs[inpt_name] = type_
            else:
                if isinstance(inpt, Arg) and inpt.type is ty.Any:
                    inpt.type = type_
    else:
        inputs = input_types
    for inpt_name, default in input_defaults.items():
        inpt = inputs[inpt_name]
        if isinstance(inpt, arg_type):
            if inpt.mandatory:
                inpt.default = default
        elif inspect.isclass(inpt) or ty.get_origin(inpt):
            inputs[inpt_name] = arg_type(type=inpt, default=default)
        elif isinstance(inpt, dict):
            inputs[inpt_name] = arg_type(**inpt)
        else:
            raise ValueError(
                f"Unrecognised input type ({inpt}) for input {inpt_name} with default "
                f"value {default}"
            )
    return_type = type_hints.get("return", ty.Any)
    if outputs:
        if len(outputs) > 1:
            if return_type is not ty.Any:
                if ty.get_origin(return_type) is not tuple:
                    raise ValueError(
                        f"Multiple outputs specified ({outputs}) but non-tuple "
                        f"return value {return_type}"
                    )
                return_types = ty.get_args(return_type)
                if len(return_types) != len(outputs):
                    raise ValueError(
                        f"Length of the outputs ({outputs}) does not match that "
                        f"of the return types ({return_types})"
                    )
                output_types = dict(zip(outputs, return_types))
            else:
                output_types = {o: ty.Any for o in outputs}
            if isinstance(outputs, dict):
                for output_name, output in outputs.items():
                    if isinstance(output, Out) and output.type is ty.Any:
                        output.type = output_types[output_name]
            else:
                outputs = output_types
        else:
            if isinstance(outputs, dict):
                output_name, output = next(iter(outputs.items()))
            elif isinstance(outputs, list):
                output_name = outputs[0]
                output = ty.Any
            if isinstance(output, Out):
                if output.type is ty.Any:
                    output.type = return_type
            elif output is ty.Any:
                output = return_type
            outputs = {output_name: output}
    elif outputs == [] or return_type in (None, type(None)):
        outputs = {}
    else:
        outputs = {"out": return_type}
    return inputs, outputs


def parse_doc_string(doc_str: str) -> tuple[dict[str, str], dict[str, str] | list[str]]:
    """Parse the docstring to pull out the description of the parameters/args and returns

    Parameters
    -----------
    doc_string
        the doc string to parse

    Returns
    -------
    input_helps
        the documentation for each of the parameter/args of the class/function
    output_helps
        the documentation for each of the return values of the class function, if no
        names are provided then the help strings are returned as a list
    """
    input_helps = {}
    output_helps = {}
    if doc_str is None:
        return input_helps, output_helps
    for param, param_help in re.findall(r":param (\w+): (.*)", doc_str):
        input_helps[param] = param_help
    for return_val, return_help in re.findall(r":return (\w+): (.*)", doc_str):
        output_helps[return_val] = return_help
    google_args_match = re.match(
        r"(?:.*\n)?\s*Args:\n(.*)", doc_str, flags=re.DOTALL | re.MULTILINE
    )
    google_returns_match = re.match(
        r"(?:.*\n)?\s*Returns:\n(.*)", doc_str, flags=re.DOTALL | re.MULTILINE
    )
    if google_args_match:
        args_str = google_args_match.group(1)
        for arg_str in split_block(args_str):
            arg_name, arg_help = arg_str.split(":", maxsplit=1)
            arg_name = arg_name.strip()
            arg_help = white_space_re.sub(" ", arg_help).strip()
            input_helps[arg_name] = arg_help
    if google_returns_match:
        returns_str = google_returns_match.group(1)
        for return_str in split_block(returns_str):
            return_name, return_help = return_str.split(":", maxsplit=1)
            return_name = return_name.strip()
            return_help = white_space_re.sub(" ", return_help).strip()
            output_helps[return_name] = return_help
    numpy_args_match = re.match(
        r"(?:.*\n)?\s+Parameters\n\s*----------\s*\n(.*)",
        doc_str,
        flags=re.DOTALL | re.MULTILINE,
    )
    numpy_returns_match = re.match(
        r"(?:.*\n)?\s+Returns\n\s*-------\s*\n(.*)",
        doc_str,
        flags=re.DOTALL | re.MULTILINE,
    )
    if numpy_args_match:
        args_str = numpy_args_match.group(1)
        for arg_str in split_block(args_str):
            arg_decl, arg_help = arg_str.split("\n", maxsplit=1)
            arg_name = arg_decl.split(":")[0].strip()
            arg_help = white_space_re.sub(" ", arg_help).strip()
            input_helps[arg_name] = arg_help
    if numpy_returns_match:
        returns_str = numpy_returns_match.group(1)
        for return_str in split_block(returns_str):
            return_decl, return_help = return_str.split("\n", maxsplit=1)
            return_name = return_decl.split(":")[0].strip()
            return_help = white_space_re.sub(" ", return_help).strip()
            output_helps[return_name] = return_help
    return input_helps, output_helps


def split_block(string: str) -> ty.Generator[str, None, None]:
    """Split a block of text into groups lines"""
    indent_re = re.compile(r"^\s*")
    leading_indent = indent_re.match(string).group()
    leading_indent_len = len(leading_indent)
    block = ""
    for line in string.split("\n"):
        if not line.strip():
            break
        indent_len = len(indent_re.match(line).group())
        if block and indent_len == leading_indent_len:
            yield block.strip()
            block = ""
        block += line + "\n"
        if indent_len < leading_indent_len:
            raise ValueError(
                f"Indentation block is not consistent in docstring:\n{string}"
            )
    if block:
        yield block.strip()


def check_explicit_fields_are_none(klass, inputs, outputs):
    if inputs is not None:
        raise ValueError(
            f"inputs should not be provided to `python.task` ({inputs}) "
            f"explicitly when decorated a class ({klass})"
        )
    if outputs is not None:
        raise ValueError(
            f"outputs should not be provided to `python.task` ({outputs}) "
            f"explicitly when decorated a class ({klass})"
        )


def sanitize_xor(
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]],
) -> set[frozenset[str]]:
    """Convert a list of xor sets into a set of frozensets"""
    # Convert a single xor set into a set of xor sets
    if not xor:
        xor = frozenset()
    elif all(isinstance(x, str) or x is None for x in xor):
        xor = frozenset([frozenset(xor)])
    else:
        xor = frozenset(frozenset(x) for x in xor)
    return xor


def extract_fields_from_class(
    spec_type: type["Task"],
    outputs_type: type["Outputs"],
    klass: type,
    arg_type: type[Arg],
    out_type: type[Out],
    auto_attribs: bool,
    skip_fields: ty.Iterable[str] = (),
) -> tuple[dict[str, Arg], dict[str, Out]]:
    """Extract the input and output fields from an existing class

    Parameters
    ----------
    klass : type
        The class to extract the fields from
    arg_type : type
        The type of the input fields
    out_type : type
        The type of the output fields
    auto_attribs : bool
        Whether to assume that all attribute annotations should be interpreted as
        fields or not
    skip_fields : Iterable[str], optional
        The names of attributes to skip when extracting the fields, by default ()

    Returns
    -------
    inputs : dict[str, Arg]
        The input fields extracted from the class
    outputs : dict[str, Out]
        The output fields extracted from the class
    """

    input_helps, _ = parse_doc_string(klass.__doc__)

    def extract_fields(klass, field_type, auto_attribs, helps) -> dict[str, Field]:
        """Get the fields from a class"""
        fields_dict = {}
        # Get fields defined in base classes if present
        for field in get_fields(klass):
            if field.name not in skip_fields:
                fields_dict[field.name] = field
        type_hints = ty.get_type_hints(klass)
        for atr_name in dir(klass):
            if (
                atr_name == "Outputs"
                or atr_name in skip_fields
                or atr_name.startswith("__")
            ):
                continue
            try:
                atr = getattr(klass, atr_name)
            except Exception:
                continue
            if isinstance(atr, Field):
                atr.name = atr_name
                fields_dict[atr_name] = atr
                if atr_name in type_hints:
                    atr.type = type_hints[atr_name]
                if not atr.help:
                    atr.help = helps.get(atr_name, "")
            elif atr_name in type_hints:
                if atr_name.startswith("_"):
                    continue
                if atr_name in fields_dict:
                    fields_dict[atr_name].type = type_hints[atr_name]
                elif auto_attribs:
                    fields_dict[atr_name] = field_type(
                        name=atr_name,
                        type=type_hints[atr_name],
                        default=atr,
                        help=helps.get(atr_name, ""),
                    )
        if auto_attribs:
            for atr_name, type_ in type_hints.items():
                if atr_name.startswith("_") or atr_name in skip_fields:
                    continue
                if atr_name not in list(fields_dict) + ["Outputs"]:
                    fields_dict[atr_name] = field_type(
                        name=atr_name, type=type_, help=helps.get(atr_name, "")
                    )
        return fields_dict

    if not issubclass(klass, spec_type):
        raise ValueError(
            f"When using the canonical form for {spec_type.__module__.split('.')[-1]} "
            f"tasks, {klass} must inherit from {spec_type}"
        )

    inputs = extract_fields(klass, arg_type, auto_attribs, input_helps)

    try:
        outputs_klass = klass.Outputs
    except AttributeError:
        raise AttributeError(
            f"Nested Outputs class not found in {klass.__name__}"
        ) from None
    if not issubclass(outputs_klass, outputs_type):
        raise ValueError(
            f"When using the canonical form for {outputs_type.__module__.split('.')[-1]} "
            f"task outputs {outputs_klass}, you must inherit from {outputs_type}"
        )

    output_helps, _ = parse_doc_string(outputs_klass.__doc__)
    outputs = extract_fields(outputs_klass, out_type, auto_attribs, output_helps)

    return inputs, outputs


white_space_re = re.compile(r"\s+")



import logging
import inspect
import typing as ty
from copy import copy, deepcopy
from collections import defaultdict
from typing import Self
import attrs
from pydra.compose import workflow
from pydra.compose.base import Task, Outputs
from pydra.engine.graph import DiGraph, INPUTS_NODE_NAME, OUTPUTS_NODE_NAME
from pydra.engine import state
from pydra.engine.lazy import LazyInField, LazyOutField
from pydra.utils.hash import hash_function, Cache
from pydra.engine.state import State
from pydra.engine.node import Node
from pydra.engine.hooks import (
    TaskHooks,
)
from pydra.engine.submitter import Submitter, NodeExecution
from pydra.utils.general import (
    attrs_values,
    asdict,
    get_fields,
)
from pydra.utils.typing import is_lazy
from pydra.environments.base import Environment

logger = logging.getLogger("pydra")

OutputsType = ty.TypeVar("OutputType", bound=Outputs)
WorkflowOutputsType = ty.TypeVar("OutputType", bound=workflow.Outputs)


@attrs.define(auto_attribs=False)
class Workflow(ty.Generic[WorkflowOutputsType]):
    """A workflow, constructed from a workflow task

    Parameters
    ----------
    name : str
        The name of the workflow
    inputs : Task
        The input task of the workflow
    outputs : Task
        The output task of the workflow
    """

    name: str = attrs.field()
    inputs: workflow.Task[WorkflowOutputsType] = attrs.field()
    outputs: WorkflowOutputsType = attrs.field()
    _nodes: dict[str, Node] = attrs.field(factory=dict)

    def __repr__(self):
        return f"Workflow(name={self.name!r}, defn={self.inputs!r})"

    @classmethod
    def clear_cache(
        cls, task: workflow.Task[WorkflowOutputsType] | None = None
    ) -> None:
        """Clear the cache of constructed workflows"""
        if task is None:
            cls._constructed_cache = defaultdict(lambda: defaultdict(dict))
        else:
            cls._constructed_cache[hash_function(task)] = defaultdict(dict)

    @classmethod
    def construct(
        cls,
        task: workflow.Task[WorkflowOutputsType],
        dont_cache: bool = False,
        lazy: ty.Sequence[str] = (),
    ) -> Self:
        """Construct a workflow from a task, caching the constructed worklow

        Parameters
        ----------
        task : workflow.Task
            The task of the workflow to construct
        dont_cache : bool, optional
            Whether to cache the constructed workflow, by default False
        lazy : Sequence[str], optional
            The names of the inputs to the workflow to be considered lazy even if they
            have values in the given task, by default ()
        """

        # Check the previously constructed workflows to see if a workflow has been
        # constructed for the given set of inputs, or a less-specific set (i.e. with a
        # super-set of lazy inputs), and use that if it exists

        non_lazy_vals = {
            n: v
            for n, v in attrs_values(task).items()
            if not is_lazy(v) and n not in lazy
        }
        non_lazy_keys = frozenset(non_lazy_vals)
        hash_cache = Cache()  # share the hash cache to avoid recalculations
        non_lazy_hash = hash_function(non_lazy_vals, cache=hash_cache)
        task_hash = hash_function(type(task), cache=hash_cache)
        # Check for same non-lazy inputs
        try:
            cached_tasks = cls._constructed_cache[task_hash]
        except KeyError:
            pass
        else:
            if (
                non_lazy_keys in cached_tasks
                and non_lazy_hash in cached_tasks[non_lazy_keys]
            ):
                return cached_tasks[non_lazy_keys][non_lazy_hash]
            # Check for supersets of lazy inputs
            for key_set, key_set_cache in cached_tasks.items():
                if key_set.issubset(non_lazy_keys):
                    subset_vals = {
                        k: v for k, v in non_lazy_vals.items() if k in key_set
                    }
                    subset_hash = hash_function(subset_vals, cache=hash_cache)
                    if subset_hash in key_set_cache:
                        wf = deepcopy(key_set_cache[subset_hash])
                        for key in non_lazy_keys - key_set:
                            # Set any additional non-lazy inputs that were not in the
                            # cached workflow.
                            setattr(wf.inputs, key, non_lazy_vals[key])
                        return wf

        # Initialise the outputs of the workflow
        outputs = task.Outputs(
            **{f.name: attrs.NOTHING for f in get_fields(task.Outputs)}
        )

        # Initialise the lzin fields
        lazy_spec = copy(task)
        workflow = Workflow(
            name=type(task).__name__,
            inputs=lazy_spec,
            outputs=outputs,
        )
        # Set lazy inputs to the workflow, need to do it after the workflow is initialised
        # so a back ref to the workflow can be set in the lazy field
        for field in get_fields(task):
            if field.name not in non_lazy_keys:
                setattr(
                    lazy_spec,
                    field.name,
                    LazyInField(
                        workflow=workflow,
                        field=field.name,
                        type=field.type,
                    ),
                )

        input_values = attrs_values(lazy_spec)
        constructor = input_values.pop("constructor")
        # Call the user defined constructor to set the outputs
        output_lazy_fields = constructor(**input_values)
        if all(v is attrs.NOTHING for v in asdict(outputs).values()):
            if output_lazy_fields is None:
                raise ValueError(
                    f"Constructor function for {task} returned None, must a lazy field "
                    "or a tuple of lazy fields"
                )
        else:  # Outputs are set explicitly in the outputs object
            if output_lazy_fields is not None:
                raise ValueError(
                    f"Constructor function for {task} must not return anything "
                    "if any of the outputs are already set explicitly"
                )
            if unset_outputs := [
                n for n, v in asdict(outputs).items() if v is attrs.NOTHING
            ]:
                raise ValueError(
                    f"Mandatory outputs {unset_outputs} are not set by the "
                    f"constructor of {workflow!r}"
                )
        # Check to see whether any mandatory inputs are not set
        for node in workflow.nodes:
            node._task._check_rules()
        # Check that the outputs are set correctly, either directly by the constructor
        # or via returned values that can be zipped with the output names
        if output_lazy_fields:
            if not isinstance(output_lazy_fields, (list, tuple)):
                output_lazy_fields = [output_lazy_fields]
            output_fields = get_fields(task.Outputs)
            if len(output_lazy_fields) != len(output_fields):
                raise ValueError(
                    f"Expected {len(output_fields)} outputs, got "
                    f"{len(output_lazy_fields)} ({output_lazy_fields})"
                )
            for outpt, outpt_lf in zip(output_fields, output_lazy_fields):
                # Automatically combine any uncombined state arrays into a single lists
                if isinstance(outpt_lf, LazyOutField):
                    outpt_lf._type = State.combine_state_arrays(outpt_lf._type)
                setattr(outputs, outpt.name, outpt_lf)
        else:
            if unset_outputs := [
                a for a, v in attrs_values(outputs).items() if v is attrs.NOTHING
            ]:
                raise ValueError(
                    f"Expected outputs {unset_outputs} to be set by the "
                    f"constructor of {workflow!r}"
                )
        if not dont_cache:
            cls._constructed_cache[task_hash][non_lazy_keys][non_lazy_hash] = workflow

        return workflow

    @classmethod
    def under_construction(cls) -> "Workflow[ty.Any]":
        """Access the under_construction variable by iterating up through the call stack."""
        frame = inspect.currentframe()
        while frame:
            # Find the frame where the construct method was called
            if (
                frame.f_code.co_name == "construct"
                and frame.f_locals.get("cls") is cls
                and "workflow" in frame.f_locals
            ):
                return frame.f_locals["workflow"]  # local var "workflow" in construct
            frame = frame.f_back
        raise RuntimeError(
            "No workflow is currently under construction (i.e. did not find a "
            "`Workflow.construct` in the current call stack"
        )

    def add(
        self,
        task: Task[OutputsType],
        name: str | None = None,
        environment: Environment | None = None,
        hooks: TaskHooks | None = None,
    ) -> OutputsType:
        """Add a node to the workflow

        Parameters
        ----------
        task_spec : Task
            The task of the job to add to the workflow as a node
        name : str, optional
            The name of the node, by default it will be the name of the task
            class
        environment : Environment, optional
            The environment to run the job in, such as the Docker or Singularity container,
            by default it will be the "native"
        hooks : TaskHooks, optional
            The hooks to run before or after the job, by default no hooks will be run

        Returns
        -------
        OutputType
            The outputs of the node
        """
        from pydra.environments import native

        if name is None:
            name = type(task).__name__
        if name in self._nodes:
            raise ValueError(f"Node with name {name!r} already exists in the workflow")
        if (
            environment
            and not isinstance(environment, native.Environment)
            and task._task_type() != "shell"
        ):
            raise ValueError(
                "Environments can only be used with 'shell' tasks not "
                f"{task._task_type()!r} tasks ({task!r})"
            )
        node = Node[OutputsType](
            name=name,
            task=task,
            workflow=self,
            environment=environment,
            hooks=hooks,
        )
        self._nodes[name] = node
        return node.lzout

    def __getitem__(self, key: str) -> Node:
        return self._nodes[key]

    @property
    def nodes(self) -> ty.Iterable[Node]:
        return self._nodes.values()

    @property
    def node_names(self) -> list[str]:
        return list(self._nodes)

    # Used to cache the constructed workflows by their hashed input values
    _constructed_cache: dict[
        str, dict[frozenset[str], dict[str, "Workflow[ty.Any]"]]
    ] = defaultdict(lambda: defaultdict(dict))

    def execution_graph(self, submitter: "Submitter") -> DiGraph:
        from pydra.engine.submitter import NodeExecution

        exec_nodes = [NodeExecution(n, submitter, workflow=self) for n in self.nodes]
        graph = self._create_graph(exec_nodes)
        # Set the graph attribute of the nodes so lazy fields can be resolved as tasks
        # are created
        for node in exec_nodes:
            node.graph = graph
        return graph

    def graph(self, detailed: bool = False) -> DiGraph:
        return self._create_graph(self.nodes, detailed=detailed)

    def _create_graph(
        self, nodes: "list[Node | NodeExecution]", detailed: bool = False
    ) -> DiGraph:
        """
        Connects a particular job to existing nodes in the workflow.

        Parameters
        ----------
        detailed : bool
            If True, `add_edges_description` is run a detailed descriptions of the
            connections (input/output fields names)
        node_klass : type, optional
            The class to use for the nodes in the workflow. If provided the node is
            wrapped by an instance of the class, if None the node is added as is,
            by default None

        Returns
        -------
        DiGraph
            The graph of the workflow
        """
        graph: DiGraph = DiGraph(name=self.name)
        for node in nodes:
            graph.add_nodes(node)
        # TODO: create connection is run twice
        for node in nodes:
            other_states = {}
            for field in get_fields(node._task):
                lf = node._task[field.name]
                if isinstance(lf, LazyOutField):
                    # adding an edge to the graph if job id expecting output from a different job

                    # checking if the connection is already in the graph
                    if (graph.node(lf._node.name), node) not in graph.edges:
                        graph.add_edges((graph.node(lf._node.name), node))
                    if detailed:
                        graph.add_edges_description(
                            (node.name, field.name, lf._node.name, lf._field)
                        )
                    logger.debug("Connecting %s to %s", lf._node.name, node.name)
                    # adding a state from the previous job to other_states
                    if (
                        graph.node(lf._node.name).state
                        and graph.node(lf._node.name).state.splitter_rpn_final
                    ):
                        # variables that are part of inner splitters should be
                        # treated as a containers
                        if (
                            node.state
                            and f"{node.name}.{field.name}"
                            in node.state._current_splitter_rpn
                        ):
                            node.state._inner_container_ndim[
                                f"{node.name}.{field.name}"
                            ] = 1
                        # adding task_name: (job.state, [a field from the connection]
                        if lf._node.name not in other_states:
                            other_states[lf._node.name] = (
                                graph.node(lf._node.name).state,
                                [field.name],
                            )
                        else:
                            # if the job already exist in other_state,
                            # additional field name should be added to the list of fields
                            other_states[lf._node.name][1].append(field.name)
                elif (
                    isinstance(lf, LazyInField) and detailed
                ):  # LazyField with the wf input
                    # connections with wf input should be added to the detailed graph description
                    graph.add_edges_description(
                        (node.name, field.name, INPUTS_NODE_NAME, lf._field)
                    )

            # if job has connections state has to be recalculated
            if other_states:
                if hasattr(node, "fut_combiner"):
                    combiner = node.fut_combiner
                else:
                    combiner = None

                if node.state:
                    node.state.update_connections(
                        new_other_states=other_states, new_combiner=combiner
                    )
                else:
                    node.state = state.State(
                        node.name,
                        splitter=None,
                        other_states=other_states,
                        combiner=combiner,
                    )
        if detailed:
            lf: LazyOutField
            for outpt_name, lf in attrs_values(self.outputs).items():
                graph.add_edges_description(
                    (OUTPUTS_NODE_NAME, outpt_name, lf._node.name, lf._field)
                )
        return graph



import typing as ty
import os
from copy import copy
import attrs
import subprocess as sp
from pathlib import Path
import logging
from fileformats.generic import FileSet
from pydra.compose import shell
from pydra.utils.general import get_fields, get_plugin_classes
from pydra.utils.typing import TypeParser
import pydra.environments

logger = logging.getLogger("pydra")

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


@attrs.define
class Environment:
    """
    Base class for environments that are used to execute tasks.
    Right now it is assumed that the environment, including container images,
    are available and are not removed at the end
    TODO: add setup and teardown methods
    """

    def setup(self):
        pass

    def execute(self, job: "Job[shell.Task]") -> dict[str, ty.Any]:
        """
        Execute the job in the environment.

        Parameters
        ----------
        job : TaskBase
            the job to execute

        Returns
        -------
        output: dict[str, Any]
            Output of the job.
        """
        raise NotImplementedError

    def teardown(self):
        pass

    @classmethod
    def available_plugins(cls) -> ty.Dict[str, ty.Type["Environment"]]:
        """Return all installed worker types"""
        return get_plugin_classes(pydra.environments, "Environment")

    @classmethod
    def plugin(cls, plugin_name: str) -> ty.Type["Environment"]:
        """Return a worker class by name."""
        try:
            return cls.available_plugins()[plugin_name.replace("-", "_")]
        except KeyError:
            raise ValueError(
                f"No environment matches {plugin_name!r}, check if there is a "
                f"plugin package called 'pydra-environments-{plugin_name}' that needs to be "
                "installed."
            )

    @classmethod
    def plugin_name(cls) -> str:
        """Return the name of the plugin."""
        try:
            plugin_name = cls._plugin_name
        except AttributeError:
            parts = cls.__module__.split(".")
            if parts[:-1] != ["pydra", "environments"]:
                raise ValueError(
                    f"Cannot infer plugin name of Environment (({cls}) from module path, "
                    f"as it isn't installed within `pydra.environments` ({cls.__module__}). "
                    "Please set the `_plugin_name` attribute on the class explicitly."
                )
            plugin_name = parts[-1]
        return plugin_name.replace("_", "-")


def split_if_str(s) -> list[str]:
    if isinstance(s, str):
        return s.split()
    elif not isinstance(s, list):
        return list(s)
    return s


@attrs.define
class Container(Environment):
    """
    Base class for container environments used by Docker and Singularity.

    Parameters
    ----------
    image : str
        Name of the container image
    tag : str
        Tag of the container image
    root : str
        Base path for mounting host directories into the container
    xargs : Union[str, List[str]]
        Extra arguments to be passed to the container
    """

    image: str
    tag: str = "latest"
    root: str = "/mnt/pydra"
    xargs: list[str] = attrs.field(factory=list, converter=split_if_str)

    def bind(self, loc, mode="ro"):
        loc_abs = Path(loc).absolute()
        return f"{loc_abs}:{self.root}{loc_abs}:{mode}"

    def get_bindings(
        self, job: "Job", root: str | None = None
    ) -> tuple[dict[str, tuple[str, str]], dict[str, tuple[Path, ...]]]:
        """Return bindings necessary to run job in an alternative root.

        This is primarily intended for contexts when a job is going
        to be run in a container with mounted volumes.

        Arguments
        ---------
        root: str, optional


        Returns
        -------
        bindings: dict
          Mapping from paths in the host environment to the target environment
        """

        bindings: dict[str, tuple[str, str]] = {}
        value_updates: dict[str, tuple[Path, ...]] = {}
        if root is None:
            return bindings
        fld: shell.arg
        for fld in get_fields(job.task):
            if TypeParser.contains_type(FileSet, fld.type):
                value: FileSet | None = job.inputs[fld.name]
                if not value:
                    continue

                copy_file = fld.copy_mode == FileSet.CopyMode.copy

                def map_path(fileset: os.PathLike | FileSet) -> Path:
                    host_path, env_path = fileset.parent, Path(
                        f"{root}{fileset.parent}"
                    )

                    # Default to mounting paths as read-only, but respect existing modes
                    bindings[host_path] = (
                        env_path,
                        "rw" if copy_file or isinstance(fld, shell.outarg) else "ro",
                    )
                    return (
                        env_path / fileset.name
                        if isinstance(fileset, os.PathLike)
                        else tuple(env_path / rel for rel in fileset.relative_fspaths)
                    )

                # Provide updated in-container paths to the command to be run. If a
                # fs-object, which resolves to a single path, just pass in the name of
                # that path relative to the location in the mount point in the container.
                # If it is a more complex file-set with multiple paths, then it is converted
                # into a tuple of paths relative to the base of the fileset.
                if TypeParser.matches(value, os.PathLike | FileSet):
                    value_updates[fld.name] = map_path(value)
                elif TypeParser.matches(value, ty.Sequence[FileSet | os.PathLike]):
                    mapped_value = []
                    for val in value:
                        mapped_val = map_path(val)
                        if isinstance(mapped_val, tuple):
                            mapped_value.extend(mapped_val)
                        else:
                            mapped_value.append(mapped_val)
                    value_updates[fld.name] = mapped_value
                else:
                    logger.debug(
                        "No support for generating bindings for %s types " "(%s)",
                        type(value),
                        value,
                    )

        # Add the cache directory to the list of mounts
        bindings[job.cache_root] = (
            f"{self.root.rstrip('/')}{job.cache_root.absolute()}",
            "rw",
        )

        # Update values with the new paths
        values = copy(job.inputs)
        values.update(value_updates)

        return bindings, values


def execute(cmd, strip=False):
    """
    Run the event loop with coroutine.

    Uses :func:`read_and_display_async` unless a loop is
    already running, in which case :func:`read_and_display`
    is used.

    Parameters
    ----------
    cmd : :obj:`list` or :obj:`tuple`
        The command line to be executed.
    strip : :obj:`bool`
        TODO

    """
    rc, stdout, stderr = read_and_display(*cmd, strip=strip)
    """
    loop = get_open_loop()
    if loop.is_running():
        rc, stdout, stderr = read_and_display(*cmd, strip=strip)
    else:
        rc, stdout, stderr = loop.run_until_complete(
            read_and_display_async(*cmd, strip=strip)
        )
    """
    return rc, stdout, stderr


def read_and_display(*cmd, strip=False, hide_display=False):
    """Capture a process' standard output."""
    try:
        process = sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    except Exception:
        # TODO editing some tracing?
        raise

    if strip:
        return (
            process.returncode,
            process.stdout.decode("utf-8").strip(),
            process.stderr.decode("utf-8"),
        )
    else:
        return (
            process.returncode,
            process.stdout.decode("utf-8"),
            process.stderr.decode("utf-8"),
        )



import typing as ty
import enum
from typing import Self
import attrs.validators
from attrs.converters import default_if_none
from fileformats.core import to_mime
from fileformats.generic import File, FileSet
from pydra.utils.typing import (
    TypeParser,
    is_truthy_falsy,
    is_type,
    is_union,
)
from pydra.utils.general import get_fields, wrap_text
import attrs

if ty.TYPE_CHECKING:
    from .task import Task


class _Empty(enum.Enum):

    NO_DEFAULT = enum.auto()

    def __repr__(self):
        return "NO_DEFAULT"

    def __bool__(self):
        return False


NO_DEFAULT = _Empty.NO_DEFAULT  # To provide a blank placeholder for the default field


def convert_default_value(value: ty.Any, self_: "Field") -> ty.Any:
    """Ensure the default value has been coerced into the correct type"""
    if value is NO_DEFAULT or isinstance(value, attrs.Factory):
        return value
    if self_.type is ty.Callable and isinstance(value, ty.Callable):
        return value
    if isinstance(self_, Out) and TypeParser.contains_type(FileSet, self_.type):
        return value
    return TypeParser[self_.type](self_.type, label=self_.name)(value)


def allowed_values_converter(value: ty.Iterable[str] | None) -> list[str] | None:
    """Ensure the allowed_values field is a list of strings or None"""
    if value is None:
        return None
    return list(value)


@attrs.define
class Requirement:
    """Define a requirement for a task input field

    Parameters
    ----------
    name : str
        The name of the input field that is required
    allowed_values : list[str], optional
        The allowed values for the input field that is required, if not provided any
        value is allowed
    """

    name: str
    allowed_values: list[str] | None = attrs.field(
        default=None, converter=allowed_values_converter
    )

    def satisfied(self, inputs: "Task") -> bool:
        """Check if the requirement is satisfied by the inputs"""
        value = getattr(inputs, self.name)
        field = {f.name: f for f in get_fields(inputs)}[self.name]
        if value is None or field.type is bool and value is False:
            return False
        if self.allowed_values is None:
            return True
        return value in self.allowed_values

    @classmethod
    def parse(cls, value: ty.Any) -> Self:
        if isinstance(value, Requirement):
            return value
        elif isinstance(value, str):
            return Requirement(value)
        elif isinstance(value, tuple):
            name, allowed_values = value
            if isinstance(allowed_values, str) or not isinstance(
                allowed_values, ty.Collection
            ):
                raise ValueError(
                    f"allowed_values must be a collection of strings, not {allowed_values}"
                )
            return Requirement(name, allowed_values)
        else:
            raise ValueError(
                f"Requirements must be a input field name, a tuple of an input field "
                f"name and allowed values or a Requirement object, not {value!r}"
            )

    def __str__(self):
        if not self.allowed_values:
            return self.name
        return f"{self.name}(" + ",".join(repr(v) for v in self.allowed_values) + ")"


def requirements_converter(value: ty.Any) -> list[Requirement]:
    """Ensure the requires field is a list of Requirement objects"""
    if isinstance(value, Requirement):
        return [value]
    elif isinstance(value, (str, tuple)):
        try:
            return [Requirement.parse(value)]
        except ValueError as e:
            e.add_note(
                f"Parsing requirements specification {value!r} as a single requirement"
            )
            raise e
    try:
        return [Requirement.parse(v) for v in value]
    except ValueError as e:
        e.add_note(
            f"Parsing requirements specification {value!r} as a set of concurrent "
            "requirements (i.e. logical AND)"
        )
        raise e


@attrs.define
class RequirementSet:
    """Define a set of requirements for a task input field, all of which must be satisfied"""

    requirements: list[Requirement] = attrs.field(
        factory=list,
        converter=requirements_converter,
    )

    def satisfied(self, inputs: "Task") -> bool:
        """Check if all the requirements are satisfied by the inputs"""
        return all(req.satisfied(inputs) for req in self.requirements)

    def __str__(self):
        if len(self.requirements) == 1:
            return str(self.requirements[0])
        return "+".join(str(r) for r in self.requirements)

    def __iter__(self):
        return iter(self.requirements)

    def __iadd__(self, other: "RequirementSet | list[Requirement]") -> "RequirementSet":
        self.requirements.extend(requirements_converter(other))
        return self


def requires_converter(
    value: (
        str
        | ty.Collection[
            Requirement | str | ty.Collection[str | tuple[str, ty.Collection[ty.Any]]]
        ]
    ),
) -> list[RequirementSet]:
    """Ensure the requires field is a tuple of tuples"""
    if isinstance(value, (str, tuple, Requirement)):
        try:
            return [RequirementSet(value)]
        except ValueError as e:
            e.add_note(
                f"Parsing requirements set specification {value!r} as a single requirement set"
            )
            raise e
    try:
        return [RequirementSet(v) for v in value]
    except ValueError as e:
        e.add_note(
            f"Parsing requirements set specification {value!r} as a set of alternative "
            "requirements (i.e. logical OR)"
        )
        raise e


@attrs.define(kw_only=True)
class Field:
    """Base class for input and output fields to tasks

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
    type: type, optional
        The type of the field, by default it is Any
        from name to field, by default it is None
    default : Any, optional
        the default value for the field, by default it is NO_DEFAULT
    help: str, optional
        A short description of the input field.
    requires: str | list[str | list[str] | Requirement], optional
        The input fields that are required to be provided, along with the optional allowed
        values, that are required together with the field. Can be provided
        as a single name, a collection of names, a collection of collections of names,
        or a collection of collection of name/allowed values pairs.
    converter: callable, optional
        The converter for the field passed through to the attrs.field, by default it is None
    validator: callable | iterable[callable], optional
        The validator(s) for the field passed through to the attrs.field, by default it is None
    hash_eq: bool, optional
        Whether to use the hash of the value for equality comparison, by default it is False
    """

    name: str | None = None
    type: ty.Type[ty.Any] = attrs.field(
        validator=is_type, default=ty.Any, converter=default_if_none(ty.Any)
    )
    default: ty.Any = attrs.field(
        default=NO_DEFAULT,
        converter=attrs.Converter(convert_default_value, takes_self=True),
    )
    help: str = ""
    requires: list[RequirementSet] = attrs.field(
        factory=list, converter=requires_converter
    )
    converter: ty.Callable[..., ty.Any] | None = None
    validator: ty.Callable[..., bool] | None = None
    hash_eq: bool = False

    def requirements_satisfied(self, inputs: "Task") -> bool:
        """Check if all the requirements are satisfied by the inputs"""
        return any(req.satisfied(inputs) for req in self.requires)

    @property
    def mandatory(self):
        return self.default is NO_DEFAULT

    @requires.validator
    def _requires_validator(self, _, value):
        if value and not is_truthy_falsy(self.type):
            raise ValueError(
                f"Fields with requirements must be of optional (i.e. in union "
                f"with None) or truthy/falsy type, not type {self.type} ({self!r})"
            )

    def markdown_listing(
        self, line_width: int = 79, help_indent: int = 4, **kwargs
    ) -> str:
        """Get the listing for the field in markdown-like format

        Parameters
        ----------
        line_width: int
            The maximum line width for the output, by default it is 79
        help_indent: int
            The indentation for the help text, by default it is 4

        Returns
        -------
        str
            The listing for the field in markdown-like format
        """

        def type_to_str(type_: ty.Type[ty.Any]) -> str:
            if type_ is type(None):
                return "None"
            if is_union(type_):
                return " | ".join(
                    type_to_str(t) for t in ty.get_args(type_) if t is not None
                )
            try:
                type_str = to_mime(type_, official=False)
            except Exception:
                if origin := ty.get_origin(type_):
                    type_str = f"{origin.__name__}[{', '.join(map(type_to_str, ty.get_args(type_)))}]"
                else:
                    try:
                        type_str = type_.__name__
                    except AttributeError:
                        type_str = str(type_)
            return type_str

        s = f"- {self.name}: {type_to_str(self.type)}"
        if isinstance(self.default, attrs.Factory):
            s += f"; default-factory = {self.default.factory.__name__}()"
        elif callable(self.default):
            s += f"; default = {self.default.__name__}()"
        elif not self.mandatory:
            s += f"; default = {self.default!r}"
        if self._additional_descriptors(**kwargs):
            s += f" ({', '.join(self._additional_descriptors(**kwargs))})"
        if self.help:
            s += f"\n{wrap_text(self.help, width=line_width, indent_size=help_indent)}"
        return s

    def _additional_descriptors(self, **kwargs) -> list[str]:
        """Get additional descriptors for the field"""
        return []

    def __lt__(self, other: "Field") -> bool:
        """Compare two fields based on their position"""
        return self.name < other.name


@attrs.define(kw_only=True)
class Arg(Field):
    """Base class for input fields of tasks

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    type: type, optional
        The type of the field, by default it is Any
    default : Any, optional
        the default value for the field, by default it is NO_DEFAULT
    help: str
        A short description of the input field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    allowed_values: Sequence, optional
        List of allowed values for the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    """

    allowed_values: frozenset = attrs.field(factory=frozenset, converter=frozenset)
    copy_mode: File.CopyMode = File.CopyMode.any
    copy_collation: File.CopyCollation = File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition = File.ExtensionDecomposition.single
    readonly: bool = False

    def _additional_descriptors(self, **kwargs) -> list[str]:
        """Get additional descriptors for the field"""
        descriptors = super()._additional_descriptors(**kwargs)
        if self.allowed_values:
            descriptors.append(f"allowed_values={self.allowed_values}")
        return descriptors


@attrs.define(kw_only=True, slots=False)
class Out(Field):
    """Base class for output fields of tasks

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    type: type, optional
        The type of the field, by default it is Any
    default : Any, optional
        the default value for the field, by default it is NO_DEFAULT
    help: str, optional
        A short description of the input field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    converter: callable, optional
        The converter for the field passed through to the attrs.field, by default it is None
    validator: callable | iterable[callable], optional
        The validator(s) for the field passed through to the attrs.field, by default it is None
    """

    pass



"""Job I/O definitions."""

from pathlib import Path
import typing as ty
import attrs
import pickle
import time
import os
import cloudpickle as cp
import getpass
from time import strftime
from filelock import SoftFileLock
from fileformats.generic import FileSet
from pydra.utils.general import (
    attrs_values,
    attrs_fields,
    is_workflow,
)
from pydra.utils.typing import copy_nested_files
from pydra.compose import workflow, base

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


TaskType = ty.TypeVar("TaskType", bound=base.Task)
OutputsType = ty.TypeVar("OutputsType", bound=base.Outputs)


@attrs.define(kw_only=True)
class Runtime:
    """Represent run time metadata."""

    rss_peak_gb: ty.Optional[float] = None
    """Peak in consumption of physical RAM."""
    vms_peak_gb: ty.Optional[float] = None
    """Peak in consumption of virtual memory."""
    cpu_peak_percent: ty.Optional[float] = None
    """Peak in cpu consumption."""


@attrs.define(kw_only=True)
class Result(ty.Generic[OutputsType]):
    """Metadata regarding the outputs of processing."""

    cache_dir: Path
    outputs: OutputsType | None = None
    runtime: Runtime | None = None
    errored: bool = False
    task: base.Task[OutputsType] | None = None

    CLOUD_PICKLE_ATTRS = ("outputs", "task")

    def __getstate__(self):
        state = attrs_values(self)
        for attr in self.CLOUD_PICKLE_ATTRS:
            if state[attr] is not None:
                state[attr] = cp.dumps(state[attr])
        return state

    def __setstate__(self, state):
        for attr in self.CLOUD_PICKLE_ATTRS:
            if state[attr] is not None:
                state[attr] = cp.loads(state[attr])
        for name, val in state.items():
            setattr(self, name, val)

    def get_output_field(self, field_name):
        """Used in get_values in Workflow

        Parameters
        ----------
        field_name : `str`
            Name of field in LazyField object
        """
        if field_name == "all_":
            return attrs_values(self.outputs)
        else:
            return getattr(self.outputs, field_name)

    @property
    def errors(self):
        if self.errored:
            error_file = self.cache_dir / "_error.pklz"
            if error_file.exists():
                with open(error_file, "rb") as f:
                    return cp.load(f)
        return None

    @property
    def job(self):
        job_pkl = self.cache_dir / "_job.pklz"
        if not job_pkl.exists():
            return None
        with open(job_pkl, "rb") as f:
            return cp.load(f)

    @property
    def return_values(self):
        return_values_pkl = self.cache_dir / "_return_values.pklz"
        if not return_values_pkl.exists():
            return None
        with open(return_values_pkl, "rb") as f:
            return cp.load(f)


@attrs.define(kw_only=True)
class RuntimeSpec:
    """
    Specification for a job.

    From CWL::

        InlineJavascriptRequirement
        SchemaDefRequirement
        DockerRequirement
        SoftwareRequirement
        InitialWorkDirRequirement
        EnvVarRequirement
        ShellCommandRequirement
        ResourceRequirement

        InlineScriptRequirement

    """

    outdir: ty.Optional[str] = None
    container: ty.Optional[str] = "shell"
    network: bool = False


def load_result(
    checksum: str,
    readonly_caches: list[Path],
    retries: int = 10,
    polling_interval: float = 0.1,
) -> Result | None:
    """
    Restore a result from the cache.

    Parameters
    ----------
    checksum : :obj:`str`
        Unique identifier of the job to be loaded.
    readonly_caches : :obj:`list` of :obj:`os.pathlike`
        List of cache directories, in order of priority, where
        the checksum will be looked for.
    retries : :obj:`int`
        Number of times to retry loading the result if the file is not
        completely written.
    polling_interval : :obj:`float`
        Time to wait between retries.

    Returns
    -------
    result : :obj:`Result` | None
        The result object if found, otherwise None.

    """
    if not readonly_caches:
        return None
    # TODO: if there are issues with loading, we might need to
    # TODO: sleep and repeat loads (after checking that there are no lock files!)
    for location in readonly_caches:
        if (location / checksum).exists():
            result_file = location / checksum / "_result.pklz"
            if result_file.exists() and result_file.stat().st_size > 0:
                # Load the result file, retrying if necessary while waiting for the file
                # to be written completely.
                for _ in range(retries):
                    try:
                        with open(result_file, "rb") as fp:
                            return cp.load(fp)
                    except (pickle.UnpicklingError, EOFError):
                        # if the file is not finished writing
                        # wait and retry
                        time.sleep(polling_interval)
            return None
    return None


def save(
    task_path: Path,
    result: "Result | None" = None,
    job: "Job[TaskType] | None" = None,
    return_values: dict[str, ty.Any] | None = None,
    name_prefix: str = None,
) -> None:
    """
    Save a :class:`~pydra.compose.base.Task` object and/or results.

    Parameters
    ----------
    task_path : :obj:`Path`
        Write directory
    result : :obj:`Result`
        Result to pickle and write
    job : :class:`~pydra.compose.base.Task`
        Job to pickle and write
    return_values : :obj:`dict`
        Return values to pickle and write
    """

    if job is None and result is None:
        raise ValueError("Nothing to be saved")

    if not isinstance(task_path, Path):
        task_path = Path(task_path)
    task_path.mkdir(parents=True, exist_ok=True)
    if name_prefix is None:
        name_prefix = ""

    lockfile = task_path.parent / (task_path.name + "_save.lock")
    with SoftFileLock(lockfile):
        if result:
            if result.task and is_workflow(result.task) and result.outputs is not None:
                # copy files to the workflow directory
                result.outputs = copyfile_workflow(
                    wf_path=task_path, outputs=result.outputs
                )
            with (task_path / f"{name_prefix}_result.pklz").open("wb") as fp:
                cp.dump(result, fp)
        if job:
            with (task_path / f"{name_prefix}_job.pklz").open("wb") as fp:
                cp.dump(job, fp)
        if return_values:
            with (task_path / f"{name_prefix}_return_values.pklz").open("wb") as fp:
                cp.dump(job, fp)


def copyfile_workflow(
    wf_path: os.PathLike, outputs: workflow.Outputs
) -> workflow.Outputs:
    """if file in the wf results, the file will be copied to the workflow directory"""

    clashes_to_avoid: set[Path] = set()
    for field in attrs_fields(outputs):
        value = getattr(outputs, field.name)
        # if the field is a path or it can contain a path _copyfile_single_value is run
        # to move all files and directories to the workflow directory
        new_value = copy_nested_files(
            value,
            wf_path,
            mode=FileSet.CopyMode.hardlink_or_copy,
            clashes_to_avoid=clashes_to_avoid,
        )
        setattr(outputs, field.name, new_value)
    return outputs


def gather_runtime_info(fname):
    """
    Extract runtime information from a file.

    Parameters
    ----------
    fname : :obj:`os.pathlike`
        The file containing runtime information

    Returns
    -------
    runtime : :obj:`Runtime`
        A runtime object containing the collected information.

    """

    runtime = Runtime(rss_peak_gb=None, vms_peak_gb=None, cpu_peak_percent=None)

    # Read .prof file in and set runtime values
    data = [
        [float(el) for el in line.strip().split(",")]
        for line in Path(fname).read_text().splitlines()
    ]
    if data:
        runtime.rss_peak_gb = max([val[2] for val in data]) / 1024
        runtime.vms_peak_gb = max([val[3] for val in data]) / 1024
        runtime.cpu_peak_percent = max([val[1] for val in data])

    """
    runtime.prof_dict = {
        'time': vals[:, 0].tolist(),
        'cpus': vals[:, 1].tolist(),
        'rss_GiB': (vals[:, 2] / 1024).tolist(),
        'vms_GiB': (vals[:, 3] / 1024).tolist(),
    }
    """
    return runtime


def create_checksum(name, inputs):
    """
    Generate a checksum name for a given combination of job name and inputs.

    Parameters
    ----------
    name : :obj:`str`
        Job name.
    inputs : :obj:`str`
        String of inputs.

    """
    return "-".join((name, inputs))


def record_error(error_path, error):
    """Write an error file."""

    error_message = str(error)

    resultfile = error_path / "_result.pklz"
    if not resultfile.exists():
        error_message += """\n
    When creating this error file, the results file corresponding
    to the job could not be found."""

    name_checksum = str(error_path.name)
    timeofcrash = strftime("%Y%m%d-%H%M%S")
    try:
        login_name = getpass.getuser()
    except KeyError:
        login_name = f"UID{os.getuid():d}"

    full_error = {
        "time of crash": timeofcrash,
        "login name": login_name,
        "name with checksum": name_checksum,
        "error message": error,
    }

    with (error_path / "_error.pklz").open("wb") as fp:
        cp.dump(full_error, fp)

    return error_path / "_error.pklz"



import typing as ty
import logging
from pydra.compose import shell
from pydra.environments import base

logger = logging.getLogger("pydra")

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


class Singularity(base.Container):
    """Singularity environment."""

    def execute(self, job: "Job[shell.Task]") -> dict[str, ty.Any]:
        singularity_img = f"{self.image}:{self.tag}"
        # mounting all input locations
        mounts, values = self.get_bindings(job=job, root=self.root)

        # todo adding xargsy etc
        singularity_args = [
            "singularity",
            "exec",
            *self.xargs,
        ]
        singularity_args.extend(
            " ".join(
                [f"-B {key}:{val[0]}:{val[1]}" for (key, val) in mounts.items()]
            ).split()
        )
        singularity_args.extend(
            ["--pwd", f"{self.root.rstrip('/')}{job.cache_dir.absolute()}"]
        )
        keys = ["return_code", "stdout", "stderr"]

        job.cache_dir.mkdir(exist_ok=True)
        values = base.execute(
            singularity_args
            + [singularity_img]
            + job.task._command_args(values=values),
        )
        output = dict(zip(keys, values))
        if output["return_code"]:
            if output["stderr"]:
                raise RuntimeError(output["stderr"])
            else:
                raise RuntimeError(output["stdout"])
        return output


# Alias so it can be referred to as singularity.Environment
Environment = Singularity



import shutil
import os
import pytest
import typing as ty
from click.testing import CliRunner, Result as CliResult

os.environ["NO_ET"] = "true"


def pytest_addoption(parser):
    parser.addoption("--with-dask", action="store_true", help="run all combinations")
    parser.addoption(
        "--only-worker",
        help="only run tests with provided worker",
    )


@pytest.fixture(scope="session", params=["debug", "cf"])
def worker(request):
    return request.param


def pytest_generate_tests(metafunc):
    if "any_worker" in metafunc.fixturenames:
        try:
            with_dask = metafunc.config.getoption("with_dask")
        except ValueError:
            with_dask = False
        try:
            only_worker = metafunc.config.getoption("only_worker")
        except ValueError:
            only_worker = None
        if only_worker is None:
            available_workers = ["debug", "cf"]
            if with_dask:
                available_workers.append("dask")
            if bool(shutil.which("sbatch")):
                available_workers.append("slurm")
        else:
            available_workers = [only_worker]
        # Set the available workers as a parameter to the
        # test function
        metafunc.parametrize("any_worker", available_workers)


@pytest.fixture
def cli_runner(catch_cli_exceptions: bool) -> ty.Callable[..., ty.Any]:
    def invoke(
        *args: ty.Any, catch_exceptions: bool = catch_cli_exceptions, **kwargs: ty.Any
    ) -> CliResult:
        runner = CliRunner()
        result = runner.invoke(*args, catch_exceptions=catch_exceptions, **kwargs)  # type: ignore[misc]
        return result

    return invoke


# For debugging in IDE's don't catch raised exceptions and let the IDE
# break at it
if os.getenv("_PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[ty.Any]) -> None:
        if call.excinfo is not None:
            raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[BaseException]) -> None:
        raise excinfo.value

    CATCH_CLI_EXCEPTIONS = False
else:
    CATCH_CLI_EXCEPTIONS = True


@pytest.fixture
def catch_cli_exceptions() -> bool:
    return CATCH_CLI_EXCEPTIONS


# Example VSCode launch configuration for debugging unittests
# {
#     "name": "Test Config",
#     "type": "python",
#     "request": "launch",
#     "purpose": ["debug-test"],
#     "justMyCode": false,
#     "console": "internalConsole",
#     "env": {
#         "_PYTEST_RAISE": "1"
#     },
# }



import typing as ty
import logging
from pydra.compose import shell
from pydra.environments import base

logger = logging.getLogger("pydra")

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


class Native(base.Environment):
    """
    Native environment, i.e. the tasks are executed in the current shell environment.
    """

    def execute(self, job: "Job[shell.Task]") -> dict[str, ty.Any]:
        keys = ["return_code", "stdout", "stderr"]
        cmd_args = job.task._command_args(values=job.inputs)
        values = base.execute(cmd_args)
        output = dict(zip(keys, values))
        if output["return_code"]:
            msg = f"Error running '{job.name}' job with {cmd_args}:"
            if output["stderr"]:
                msg += "\n\nstderr:\n" + output["stderr"]
            if output["stdout"]:
                msg += "\n\nstdout:\n" + output["stdout"]
            raise RuntimeError(msg)
        return output


# Alias so it can be referred to as native.Environment
Environment = Native



"""Generic object hashing dispatch"""

import sys
import os
import re
import ast
import struct
import inspect
from datetime import datetime
import typing as ty
import types
from pathlib import Path
from collections.abc import Mapping
from functools import singledispatch
from hashlib import blake2b
import logging
from typing import (
    Dict,
    Iterator,
    NewType,
    Sequence,
    Set,
)
from filelock import SoftFileLock
import attrs.exceptions
from fileformats.core.fileset import FileSet, MockMixin
from fileformats.generic import FsObject
import fileformats.core.exceptions
from pydra.utils.general import in_stdlib, user_cache_root, add_exc_note

logger = logging.getLogger("pydra")

FUNCTION_SRC_CHUNK_LEN_DEFAULT = 8192

try:
    from typing import Protocol
except ImportError:
    from typing import Protocol  # type: ignore

try:
    from typing import runtime_checkable
except ImportError:
    from typing import runtime_checkable  # type: ignore


try:
    import numpy
except ImportError:
    HAVE_NUMPY = False
else:
    HAVE_NUMPY = True

__all__ = (
    "hash_function",
    "hash_object",
    "hash_single",
    "register_serializer",
    "Hash",
    "Cache",
    "bytes_repr_mapping_contents",
    "bytes_repr_sequence_contents",
)

Hash = NewType("Hash", bytes)
CacheKey = NewType("CacheKey", ty.Tuple[ty.Hashable, ...])


def location_converter(path: ty.Union[Path, str, None]) -> Path:
    if path is None:
        path = PersistentCache.location_default()
    path = Path(path)
    try:
        path.mkdir(parents=True, exist_ok=True)
    except FileExistsError:
        raise ValueError(
            f"provided path to persistent cache {path} is a file not a directory"
        ) from None
    return path


@attrs.define
class PersistentCache:
    """Persistent cache in which to store computationally expensive hashes between nodes
    and workflow/task runs. It does this in via the `get_or_calculate_hash` method, which
    takes a locally unique key (e.g. file-system path + mtime) and a function to
    calculate the hash if it isn't present in the persistent store.

    The locally unique key is hashed (cheaply) using hashlib cryptography and this
    "local hash" is use to name the entry of the (potentially expensive) hash of the
    object itself (e.g. the contents of a file). This entry is saved as a text file
    within a user-specific cache directory (see `platformdirs.user_cache_dir`), with
    the name of the file being the "local hash" of the key and the contents of the
    file being the "globally unique hash" of the object itself.

    Parameters
    ----------
    location: Path
        the directory in which to store the hashes cache
    """

    location: Path = attrs.field(converter=location_converter)  # type: ignore[misc]
    cleanup_period: int = attrs.field()
    _hashes: ty.Dict[CacheKey, Hash] = attrs.field(factory=dict)

    # Set the location of the persistent hash cache
    LOCATION_ENV_VAR = "PYDRA_HASH_CACHE"
    CLEANUP_ENV_VAR = "PYDRA_HASH_CACHE_CLEANUP_PERIOD"

    @classmethod
    def location_default(cls):
        try:
            location = os.environ[cls.LOCATION_ENV_VAR]
        except KeyError:
            location = user_cache_root / "hashes"
        return location

    # the default needs to be an instance method
    @location.default
    def _location_default(self):
        return self.location_default()

    @cleanup_period.default
    def cleanup_period_default(self):
        return int(os.environ.get(self.CLEANUP_ENV_VAR, 30))

    def get_or_calculate_hash(self, key: CacheKey, calculate_hash: ty.Callable) -> Hash:
        """Check whether key is present in the persistent cache store and return it if so.
        Otherwise use `calculate_hash` to generate the hash and save it in the persistent
        store.

        Parameters
        ----------
        key : CacheKey
            locally unique key (e.g. to the host) used to lookup the corresponding hash
            in the persistent store
        calculate_hash : ty.Callable
            function to calculate the hash if it isn't present in the persistent store

        Returns
        -------
        Hash
            the hash corresponding to the key, which is either retrieved from the persistent
            store or calculated using `calculate_hash` if not present
        """
        try:
            return self._hashes[key]
        except KeyError:
            pass
        key_path = self.location / blake2b(str(key).encode()).hexdigest()
        with SoftFileLock(key_path.with_suffix(".lock")):
            if key_path.exists():
                return Hash(key_path.read_bytes())
            hsh = calculate_hash()
            key_path.write_bytes(hsh)
            self._hashes[key] = Hash(hsh)
        return Hash(hsh)

    def clean_up(self):
        """Cleans up old hash caches that haven't been accessed in the last 30 days"""
        now = datetime.now()
        for path in self.location.iterdir():
            if path.name.endswith(".lock"):
                continue
            days = (now - datetime.fromtimestamp(path.lstat().st_atime)).days
            if days > self.cleanup_period:
                path.unlink()

    @classmethod
    def from_path(
        cls, path: ty.Union[Path, str, "PersistentCache", None]
    ) -> "PersistentCache":
        if isinstance(path, PersistentCache):
            return path
        return PersistentCache(path)


@attrs.define
class Cache:
    """Cache for hashing objects, used to avoid infinite recursion caused by circular
    references between objects, and to store hashes of objects that have already been
    hashed to avoid recomputation.

    This concept is extended to persistent caching of hashes for certain object types,
    for which calculating the hash is a potentially expensive operation (e.g.
    File/Directory types). For these classes the `bytes_repr` override function yields a
    "locally unique cache key" (e.g. file-system path + mtime) as the first item of its
    iterator.
    """

    persistent: ty.Optional[PersistentCache] = attrs.field(
        default=None,
        converter=PersistentCache.from_path,  # type: ignore[misc]
    )
    _hashes: ty.Dict[int, Hash] = attrs.field(factory=dict)

    def __getitem__(self, object_id: int) -> Hash:
        return self._hashes[object_id]

    def __setitem__(self, object_id: int, hsh: Hash):
        self._hashes[object_id] = hsh

    def __contains__(self, object_id):
        return object_id in self._hashes


def hash_function(obj, **kwargs):
    """Generate hash of object."""
    return hash_object(obj, **kwargs).hex()


def hash_object(
    obj: object,
    cache: ty.Optional[Cache] = None,
    persistent_cache: ty.Union[PersistentCache, Path, None] = None,
) -> Hash:
    """Hash an object

    Constructs a byte string that uniquely identifies the object,
    and returns the hash of that string.

    Base Python types are implemented, including recursive lists and
    dicts. Custom types can be registered with :func:`register_serializer`.
    """
    if cache is None:
        cache = Cache(persistent=persistent_cache)
    try:
        return hash_single(obj, cache)
    except Exception as e:
        tp = type(obj)
        add_exc_note(
            e,
            (
                f"and therefore cannot hash `{obj!r}` of type "
                f"`{tp.__module__}.{tp.__name__}`. Consider implementing a "
                "specific `bytes_repr()`(see pydra.utils.hash.register_serializer) "
                "or a `__bytes_repr__()` dunder methods for this type"
            ),
        )
        raise e


def hash_single(obj: object, cache: Cache) -> Hash:
    """Single object-scoped hash

    Uses a local cache to prevent infinite recursion. This cache is unsafe
    to reuse across multiple objects, so this function should not be used directly.
    """
    objid = id(obj)
    if objid not in cache:
        # Handle recursion by putting a dummy value in the cache
        cache[objid] = Hash(b"\x00")
        bytes_it = bytes_repr(obj, cache)
        # Pop first element from the bytes_repr iterator and check whether it is a
        # "local cache key" (e.g. file-system path + mtime tuple) or the first bytes
        # chunk

        def calc_hash(first: ty.Optional[bytes] = None) -> Hash:
            """
            Calculate the hash of the object

            Parameters
            ----------
            first : ty.Optional[bytes]
                the first bytes chunk from the bytes_repr iterator, passed if the first
                chunk wasn't a local cache key
            """
            h = blake2b(digest_size=16, person=b"pydra-hash")
            # We want to use the first chunk that was popped to check for a cache-key
            # if present
            if first is not None:
                h.update(first)
            for chunk in bytes_it:  # Note that `bytes_it` is in outer scope
                h.update(chunk)
            return Hash(h.digest())

        # Read the first item of the bytes_repr iterator and check to see whether it yields
        # a "cache-key" tuple instead of a bytes chunk for the type of the object to be cached
        # (e.g. file-system path + mtime for fileformats.core.FileSet objects). If it
        # does, use that key to check the persistent cache for a precomputed hash and
        # return it if it is, otherwise calculate the hash and store it in the persistent
        # cache with that hash of that key (not to be confused with the hash of the
        # object that is saved/retrieved).
        first = next(bytes_it)
        if isinstance(first, tuple):
            tp = type(obj)
            key = (
                tp.__module__,
                tp.__name__,
            ) + first
            hsh = cache.persistent.get_or_calculate_hash(key, calc_hash)
        else:
            # If the first item is a bytes chunk (i.e. the object type doesn't have an
            # associated 'cache-key'), then simply calculate the hash of the object,
            # passing the first chunk to the `calc_hash` function so it can be included
            # in the hash calculation
            hsh = calc_hash(first=first)
        logger.debug("Hash of %s object is %s", obj, hsh)
        cache[objid] = hsh
    return cache[objid]


@runtime_checkable
class HasBytesRepr(Protocol):
    def __bytes_repr__(self, cache: Cache) -> Iterator[bytes]:
        pass  # pragma: no cover


@singledispatch
def bytes_repr(obj: object, cache: Cache) -> Iterator[bytes]:
    """Default implementation of hashing for generic objects. Single dispatch is used
    to provide hooks for class-specific implementations

    Parameters
    ----------
    obj: object
        the object to hash
    cache : Cache
        a dictionary object used to store a cache of previously cached objects to
        handle circular object references

    Yields
    -------
    bytes
        unique representation of the object in a series of bytes
    """
    cls = obj.__class__
    yield f"{cls.__module__}.{cls.__name__}:{{".encode()
    dct: Dict[str, ty.Any]
    if attrs.has(type(obj)):
        # Drop any attributes that aren't used in comparisons by default
        dct = attrs.asdict(obj, recurse=False, filter=lambda a, _: bool(a.eq))
    elif hasattr(obj, "__slots__") and obj.__slots__ is not None:
        dct = {attr: getattr(obj, attr) for attr in obj.__slots__}
    else:

        def is_special_or_method(n: str):
            return (n.startswith("__") and n.endswith("__")) or inspect.ismethod(
                getattr(obj, n)
            )

        try:
            dct = {n: v for n, v in obj.__dict__.items() if not is_special_or_method(n)}
        except AttributeError:
            dct = {n: getattr(obj, n) for n in dir(obj) if not is_special_or_method(n)}
    yield from bytes_repr_mapping_contents(dct, cache)
    yield b"}"


register_serializer = bytes_repr.register
register_serializer.__doc__ = """Register a custom serializer for a type

The generator function should yield byte strings that will be hashed
to produce the final hash. A recommended convention is to yield a
qualified type prefix (e.g. ``f"{module}.{class}"``),
followed by a colon, followed by the serialized value.

If serializing an iterable, an open and close bracket may be yielded
to identify the start and end of the iterable.

Consider using :func:`bytes_repr_mapping_contents` and
:func:`bytes_repr_sequence_contents` to serialize the contents of a mapping
or sequence. These do not include the prefix or brackets, so they can be
reused as part of a custom serializer.

As an example, the following example is the default serializer for user-defined
classes:

.. code-block:: python

    @register_serializer
    def bytes_repr(obj: object, cache: Cache) -> Iterator[bytes]:
        cls = obj.__class__
        yield f"{cls.__module__}.{cls.__name__}:{{".encode()
        yield from bytes_repr_mapping_contents(obj.__dict__, cache)
        yield b"}"

Serializers must accept a ``cache`` argument, which is a dictionary that
permits caching of hashes for recursive objects. If the hash of sub-objects
is used to create an object serialization, the :func:`hash_single` function
should be called with the same cache object.
"""


@register_serializer
def bytes_repr_dunder(obj: HasBytesRepr, cache: Cache) -> Iterator[bytes]:
    yield from obj.__bytes_repr__(cache)


@register_serializer(type(None))
@register_serializer(type(Ellipsis))
@register_serializer(bool)
@register_serializer(range)
def bytes_repr_builtin_repr(
    obj: object,
    cache: Cache,
) -> Iterator[bytes]:
    yield repr(obj).encode()


@register_serializer
def bytes_repr_slice(obj: slice, cache: Cache) -> Iterator[bytes]:
    yield b"slice("
    yield from bytes_repr_sequence_contents((obj.start, obj.stop, obj.step), cache)
    yield b")"


@register_serializer
def bytes_repr_pathlike(obj: os.PathLike, cache: Cache) -> Iterator[bytes]:
    cls = obj.__class__
    yield f"{cls.__module__}.{cls.__name__}:{os.fspath(obj)}".encode()


@register_serializer
def bytes_repr_bytes(obj: bytes, cache: Cache) -> Iterator[bytes]:
    yield f"bytes:{len(obj)}:".encode()
    yield obj


@register_serializer
def bytes_repr_str(obj: str, cache: Cache) -> Iterator[bytes]:
    val = obj.encode()
    yield f"str:{len(val)}:".encode()
    yield val


@register_serializer
def bytes_repr_int(obj: int, cache: Cache) -> Iterator[bytes]:
    try:
        # Up to 64-bit ints
        val = struct.pack("<q", obj)
        yield b"int:"
    except struct.error:
        # Big ints (old python "long")
        val = str(obj).encode()
        yield f"long:{len(val)}:".encode()
    yield val


@register_serializer
def bytes_repr_float(obj: float, cache: Cache) -> Iterator[bytes]:
    yield b"float:"
    yield struct.pack("<d", obj)


@register_serializer
def bytes_repr_complex(obj: complex, cache: Cache) -> Iterator[bytes]:
    yield b"complex:"
    yield struct.pack("<dd", obj.real, obj.imag)


@register_serializer
def bytes_repr_dict(obj: dict, cache: Cache) -> Iterator[bytes]:
    yield b"dict:{"
    yield from bytes_repr_mapping_contents(obj, cache)
    yield b"}"


@register_serializer
def bytes_repr_module(obj: types.ModuleType, cache: Cache) -> Iterator[bytes]:
    yield b"module:("
    yield hash_single(FsObject(obj.__file__), cache=cache)
    yield b")"


@register_serializer(ty._GenericAlias)
@register_serializer(ty._SpecialForm)
@register_serializer(type)
def bytes_repr_type(klass: type, cache: Cache) -> Iterator[bytes]:
    from pydra.utils.general import get_fields

    def type_location(tp: type) -> bytes:
        """Return the module and name of the type in a ASCII byte string"""
        try:
            type_name = tp.__name__
        except AttributeError:
            type_name = tp._name
        mod_path = ".".join(
            p for p in klass.__module__.split(".") if not p.startswith("_")
        )
        return f"{mod_path}.{type_name}".encode()

    yield b"type:("
    origin = ty.get_origin(klass)
    args = ty.get_args(klass)
    if origin and args:
        yield b"origin:("
        yield from bytes_repr_type(origin, cache)
        yield b"),args:("
        for arg in args:
            if isinstance(
                arg, list
            ):  # sometimes (e.g. Callable) the args of a type is a list
                yield b"list:("
                yield from (b for t in arg for b in bytes_repr_type(t, cache))
                yield b")"
            else:
                yield from bytes_repr_type(arg, cache)
        yield b")"
    else:
        if klass is Ellipsis:
            yield b"..."
        elif inspect.isclass(klass) and issubclass(klass, FileSet):
            try:
                yield b"mime-like:(" + klass.mime_like.encode() + b")"
            except fileformats.core.exceptions.FormatDefinitionError:
                yield type_location(klass)
        elif fields := get_fields(klass):
            yield b"fields:("
            yield from bytes_repr_sequence_contents(fields, cache)
            yield b")"
            if hasattr(klass, "Outputs"):
                yield b",outputs:("
                yield from bytes_repr_type(klass.Outputs, cache)
                yield b")"
        elif in_stdlib(klass):
            yield type_location(klass)
        else:
            try:
                dct = {
                    n: v for n, v in klass.__dict__.items() if not n.startswith("__")
                }
            except AttributeError:
                yield type_location(klass)
            else:
                yield b"__dict__:("
                yield from bytes_repr_mapping_contents(dct, cache)
                yield b")"
                # Include annotations
                try:
                    annotations = klass.__annotations__
                except AttributeError:
                    pass
                else:
                    yield b",annotations:("
                    yield from bytes_repr_mapping_contents(annotations, cache)
                    yield b")"
                yield b",mro:("
                yield from (
                    b for t in klass.mro()[1:-1] for b in bytes_repr_type(t, cache)
                )
                yield b")"
    yield b")"


if sys.version_info >= (3, 10):
    register_serializer(types.UnionType)(bytes_repr_type)


@register_serializer(FileSet)
def bytes_repr_fileset(
    fileset: FileSet, cache: Cache
) -> Iterator[ty.Union[CacheKey, bytes]]:
    fspaths = sorted(fileset.fspaths)
    # Yield the cache key for the fileset, which is a tuple of the file-system paths
    # and their mtime. Is used to store persistent cache of the fileset hashes
    # to avoid recomputation between calls
    yield CacheKey(
        tuple(repr(p) for p in fspaths)  # type: ignore[arg-type]
        + tuple(p.lstat().st_mtime_ns for p in fspaths)
    )
    cls = type(fileset)
    yield f"{cls.__module__}.{cls.__name__}:".encode()
    for key, chunk_iter in fileset.byte_chunks():
        yield (",'" + key + "'=").encode()
        yield from chunk_iter


# Need to disable the mtime cache key for mocked filesets. Used in doctests
@register_serializer(MockMixin)
def bytes_repr_mock_fileset(
    mock_fileset: MockMixin, cache: Cache
) -> Iterator[ty.Union[CacheKey, bytes]]:
    cls = type(mock_fileset)
    yield f"{cls.__module__}.{cls.__name__}:".encode()
    for key, _ in mock_fileset.byte_chunks():
        yield (",'" + key + "'").encode()


@register_serializer(list)
@register_serializer(tuple)
def bytes_repr_seq(obj: Sequence, cache: Cache) -> Iterator[bytes]:
    yield f"{obj.__class__.__name__}:(".encode()
    yield from bytes_repr_sequence_contents(obj, cache)
    yield b")"


@register_serializer(set)
@register_serializer(frozenset)
def bytes_repr_set(obj: Set, cache: Cache) -> Iterator[bytes]:
    yield f"{obj.__class__.__name__}:{{".encode()
    yield from bytes_repr_sequence_contents(sorted(obj), cache)
    yield b"}"


@register_serializer
def bytes_repr_code(obj: types.CodeType, cache: Cache) -> Iterator[bytes]:
    yield b"code:("
    yield from bytes_repr_sequence_contents(
        (
            obj.co_argcount,
            obj.co_posonlyargcount,
            obj.co_kwonlyargcount,
            obj.co_nlocals,
            obj.co_flags,
            obj.co_code,
            obj.co_consts,
            obj.co_names,
            obj.co_varnames,
            obj.co_freevars,
            obj.co_name,
            obj.co_cellvars,
        ),
        cache,
    )
    yield b")"


@register_serializer
def bytes_repr_function(obj: types.FunctionType, cache: Cache) -> Iterator[bytes]:
    """Serialize a function, attempting to use the AST of the source code if available
    otherwise falling back to the byte-code of the function."""
    yield b"function:("
    if in_stdlib(obj):
        yield f"{obj.__module__}.{obj.__name__}".encode()
    else:
        try:
            src = inspect.getsource(obj)
        except OSError:
            # Fallback to using the bytes representation of the code object
            yield from bytes_repr(obj.__code__, cache)
        else:

            def dump_ast(node: ast.AST) -> bytes:
                return ast.dump(
                    node, annotate_fields=False, include_attributes=False
                ).encode()

            def strip_annotations(node: ast.AST):
                """Remove annotations from function arguments."""
                if hasattr(node, "args"):
                    for arg in node.args.args:
                        arg.annotation = None
                    for arg in node.args.kwonlyargs:
                        arg.annotation = None
                    if node.args.vararg:
                        node.args.vararg.annotation = None
                    if node.args.kwarg:
                        node.args.kwarg.annotation = None

            indent = re.match(r"(\s*)", src).group(1)
            if indent:
                src = re.sub(f"^{indent}", "", src, flags=re.MULTILINE)
            try:
                func_ast = ast.parse(src).body[0]
                strip_annotations(func_ast)
                if hasattr(func_ast, "args"):
                    yield dump_ast(func_ast.args)
                if hasattr(func_ast, "body"):
                    for stmt in func_ast.body:
                        yield dump_ast(stmt)
            except SyntaxError:
                yield src.encode()
    yield b")"


def bytes_repr_mapping_contents(mapping: Mapping, cache: Cache) -> Iterator[bytes]:
    """Serialize the contents of a mapping

    Concatenates byte-serialized keys and hashed values.

    .. code-block:: python

        >>> from pydra.utils.hash import bytes_repr_mapping_contents, Cache
        >>> generator = bytes_repr_mapping_contents({"a": 1, "b": 2}, Cache())
        >>> b''.join(generator)
        b'str:1:a=...str:1:b=...'
    """
    for key in sorted(mapping):
        yield from bytes_repr(key, cache)
        yield b"="
        yield bytes(hash_single(mapping[key], cache))
        yield b","


def bytes_repr_sequence_contents(seq: Sequence, cache: Cache) -> Iterator[bytes]:
    """Serialize the contents of a sequence

    Concatenates hashed values.

    .. code-block:: python

        >>> from pydra.utils.hash import bytes_repr_sequence_contents, Cache
        >>> generator = bytes_repr_sequence_contents([1, 2], Cache())
        >>> list(generator)
        [b'\x6d...', b'\xa3...']
    """
    for val in seq:
        yield bytes(hash_single(val, cache))


if HAVE_NUMPY:

    @register_serializer(numpy.generic)
    @register_serializer(numpy.ndarray)
    def bytes_repr_numpy(obj: numpy.ndarray, cache: Cache) -> Iterator[bytes]:
        yield f"{obj.__class__.__module__}{obj.__class__.__name__}:{obj.size}:".encode()
        if obj.dtype == "object":
            yield from bytes_repr_sequence_contents(iter(obj.ravel()), cache)
        else:
            yield obj.tobytes(order="C")


NUMPY_CHUNK_LEN = 8192



"""Data structure to support :class:`~pydra.engine.workflow.Workflow` tasks."""

from copy import copy
from pathlib import Path
import typing as ty
from collections import Counter
import subprocess as sp
from pydra.utils.general import ensure_list, is_workflow


NodeType = ty.TypeVar("NodeType")

INPUTS_NODE_NAME = "__INPUTS__"
OUTPUTS_NODE_NAME = "__OUTPUTS__"


class DiGraph(ty.Generic[NodeType]):
    """A simple Directed Graph object."""

    name: str
    nodes: list[NodeType]
    edges: list[tuple[NodeType, NodeType]]

    def __init__(
        self,
        name: str | None = None,
        nodes: ty.Iterable[NodeType] | None = None,
        edges: ty.Iterable[tuple[NodeType, NodeType]] | None = None,
    ):
        """
        Initialize a directed graph.

        Parameters
        ----------
        nodes :
            Tasks are represented by the nodes of the graph.
        edges :
            Connections of inputs and outputs between tasks in
            the graph.

        """
        self.name = name
        self._nodes = []
        self.nodes = nodes
        self._edges = []
        self.edges = edges
        self._create_connections()
        self._sorted_nodes = None
        self._node_wip = []
        self._nodes_details = {}
        self._node_lookup = {}

    def copy(self):
        """
        Duplicate this graph.

        Create a copy that contains new lists and dictionaries,
        but runnable objects are the same.

        """
        cls = self.__class__
        new_graph = cls.__new__(cls)
        new_graph._nodes = self._nodes[:]
        new_graph._node_wip = self._node_wip[:]
        new_graph._edges = self._edges[:]
        if self._sorted_nodes:
            new_graph._sorted_nodes = self._sorted_nodes[:]
        else:
            new_graph._sorted_nodes = None
        new_graph.predecessors = {}
        for key, val in self.predecessors.items():
            new_graph.predecessors[key] = self.predecessors[key][:]
        new_graph.successors = {}
        for key, val in self.successors.items():
            new_graph.successors[key] = self.successors[key][:]
        return new_graph

    @property
    def nodes(self) -> list[NodeType]:
        """Get a list of the nodes currently contained in the graph."""
        return self._nodes

    @nodes.setter
    def nodes(self, nodes: ty.Iterable[NodeType]) -> None:
        if nodes:
            if duplicate_names := [
                n
                for n, c in Counter(nd.name for nd in ensure_list(nodes)).items()
                if c > 1
            ]:
                raise ValueError(
                    f"Duplicate node names found in graph: {duplicate_names}"
                )
            self._nodes = nodes

    def node(self, name: str) -> NodeType:
        """Get a node by its name, caching the lookup directory"""
        try:
            return self._node_lookup[name]
        except KeyError:
            self._node_lookup = self.nodes_names_map
            try:
                return self._node_lookup[name]
            except KeyError:
                raise KeyError(f"Node {name!r} not found in graph") from None

    def __getitem__(self, key):
        """Get a node by its name."""
        return self.node(key)

    @property
    def nodes_names_map(self) -> dict[str, NodeType]:
        """Get a map of node names to nodes."""
        return {nd.name: nd for nd in self.nodes}

    @property
    def edges(self):
        """Get a list of the links between nodes."""
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Add edges to the graph (nodes should be already set)."""
        if edges:
            edges = ensure_list(edges)
            for nd_out, nd_in in edges:
                if nd_out not in self.nodes or nd_in not in self.nodes:
                    raise Exception(
                        f"edge {(nd_out, nd_in)} can't be added to the graph"
                    )
            self._edges = edges

    @property
    def edges_names(self):
        """Get edges as pairs of the nodes they connect."""
        return [(edg[0].name, edg[1].name) for edg in self._edges]

    @property
    def nodes_details(self):
        """dictionary with details of the nodes
        for each task, there are inputs/outputs and connections
        (with input/output fields names)
        """
        # removing repeated fields from inputs and outputs
        for el in self._nodes_details.values():
            el["inputs"] = list(set(el["inputs"]))
            el["inputs"].sort()
            el["outputs"] = list(set(el["outputs"]))
            el["outputs"].sort()
        return self._nodes_details

    @property
    def sorted_nodes(self):
        """Return sorted nodes (runs sorting if needed)."""
        if self._sorted_nodes is None:
            self.sorting()
        return self._sorted_nodes

    @property
    def sorted_nodes_names(self):
        """Return a list of sorted nodes names."""
        return [nd.name for nd in self._sorted_nodes]

    def _create_connections(self):
        """Create connections between nodes."""
        self.predecessors = {}
        self.successors = {}
        for nd in self.nodes:
            self.predecessors[nd.name] = []
            self.successors[nd.name] = []

        for nd_out, nd_in in self.edges:
            self.predecessors[nd_in.name].append(nd_out)
            self.successors[nd_out.name].append(nd_in)

    def add_nodes(self, new_nodes):
        """Insert new nodes and sort the new graph."""
        self.nodes = self._nodes + ensure_list(new_nodes)
        for nd in ensure_list(new_nodes):
            self.predecessors[nd.name] = []
            self.successors[nd.name] = []
        if self._sorted_nodes is not None:
            # starting from the previous sorted list, so is faster
            self.sorting(presorted=self.sorted_nodes + ensure_list(new_nodes))

    def add_edges(self, new_edges):
        """Add new edges and sort the new graph."""
        self.edges = self._edges + ensure_list(new_edges)
        for nd_out, nd_in in ensure_list(new_edges):
            self.predecessors[nd_in.name].append(nd_out)
            self.successors[nd_out.name].append(nd_in)
        if self._sorted_nodes is not None:
            # starting from the previous sorted list, so it's faster
            self.sorting(presorted=self.sorted_nodes + [])

    def add_edges_description(self, new_edge_details):
        """adding detailed description of the connections, filling _nodes_details"""
        in_nd, in_fld, out_nd, out_fld = new_edge_details
        for key in [in_nd, out_nd]:
            self._nodes_details.setdefault(
                key, {"inputs": [], "outputs": [], "connections": []}
            )

        if (in_fld, out_nd, out_fld) not in self._nodes_details[in_nd]["connections"]:
            self._nodes_details[in_nd]["connections"].append((in_fld, out_nd, out_fld))
            self._nodes_details[in_nd]["inputs"].append(in_fld)
            self._nodes_details[out_nd]["outputs"].append(out_fld)

    def sorting(self, presorted=None):
        """
        Sort this graph.

        Sorting starts either from self.nodes or the
        previously sorted list.

        Parameters
        ----------
        presorted : :obj:`list`
            A list of previously sorted nodes.

        """
        self._sorted_nodes = []
        if presorted:
            notsorted_nodes = copy(presorted)
        else:
            notsorted_nodes = copy(self.nodes)
        predecessors = {key: copy(val) for (key, val) in self.predecessors.items()}

        # nodes that depends only on the self._nodes_wip should go first
        # soe remove them from the connections
        for nd_out in self._node_wip:
            for nd_in in self.successors[nd_out.name]:
                predecessors[nd_in.name].remove(nd_out)

        while notsorted_nodes:
            sorted_part, notsorted_nodes = self._sorting(notsorted_nodes, predecessors)
            self._sorted_nodes += sorted_part
            for nd_out in sorted_part:
                for nd_in in self.successors[nd_out.name]:
                    predecessors[nd_in.name].remove(nd_out)

    def _sorting(self, notsorted_list, predecessors):
        """
        Sort implementation.

        Adding nodes that don't have predecessors to the sorted_parts,
        returns sorted part and remaining nodes.

        """
        remaining_nodes = []
        sorted_part = []
        for nd in notsorted_list:
            if not predecessors[nd.name]:
                sorted_part.append(nd)
            else:
                remaining_nodes.append(nd)
        return sorted_part, remaining_nodes

    def remove_nodes(self, nodes, check_ready=True):
        """
        Mark nodes for removal from the graph, re-sorting if needed.

        .. important ::
            This method does not remove connections, see
            :py:meth:`~DiGraph.remove_node_connections`.
            Nodes are added to the ``_node_wip`` list, marking
            them for removal when all referring connections
            are removed.

        Parameters
        ----------
        nodes : :obj:`list`
            List of nodes to be marked for removal.
        check_ready : :obj: `bool`
            checking if the node is ready to be removed

        """
        nodes = ensure_list(nodes)
        for nd in nodes:
            if nd not in self.nodes:
                raise Exception(f"{nd} is not present in the graph")
            if self.predecessors[nd.name] and check_ready:
                raise Exception("this node shouldn't be run, has to wait")
            self.nodes.remove(nd)
            # adding the node to self._node_wip as for
            self._node_wip.append(nd)
        # if graph is sorted, the sorted list has to be updated
        if hasattr(self, "sorted_nodes"):
            if nodes == self.sorted_nodes[: len(nodes)]:
                # if the first node is removed, no need to sort again
                self._sorted_nodes = self.sorted_nodes[len(nodes) :]
            else:
                for nd in nodes:
                    self._sorted_nodes.remove(nd)
                # starting from the previous sorted list, so is faster
                self.sorting(presorted=self.sorted_nodes)
        # Reset the node lookup
        self._node_lookup = {}

    def remove_nodes_connections(self, nodes):
        """
        Remove connections between nodes.

        Also prunes the nodes from ``_node_wip``.

        Parameters
        ----------
        nodes : :obj:`list`
            List of nodes which connections are to be removed.

        """
        nodes = ensure_list(nodes)
        for nd in nodes:
            for nd_in in self.successors[nd.name]:
                self.predecessors[nd_in.name].remove(nd)
                self.edges.remove((nd, nd_in))
            self.successors.pop(nd.name)
            self.predecessors.pop(nd.name)
            self._node_wip.remove(nd)
        # Reset the node lookup
        self._node_lookup = {}

    def remove_previous_connections(self, nodes):
        """
        Remove connections that the node has with predecessors.

        Also prunes the nodes from ``_node_wip``.

        Parameters
        ----------
        nodes : :obj:`list`
            List of nodes which connections are to be removed.

        """
        nodes = ensure_list(nodes)
        for nd in nodes:
            for nd_out in self.predecessors[nd.name]:
                if nd_out.name in self.successors:
                    self.successors[nd_out.name].remove(nd)
                self.edges.remove((nd_out, nd))
            self.successors.pop(nd.name)
            self.predecessors.pop(nd.name)
            self._node_wip.remove(nd)
        # Reset the node lookup
        self._node_lookup = {}

    def _checking_successors_nodes(self, node, remove=True):
        if self.successors[node.name]:
            for nd_in in self.successors[node.name]:
                self._successors_all.append(nd_in)
                self._checking_successors_nodes(node=nd_in)
        else:
            return True

    def successors_nodes(self, node):
        """Get all the nodes that follow the node"""
        self._successors_all = []
        self._checking_successors_nodes(node=node, remove=False)
        return set(self._successors_all)

    def remove_successors_nodes(self, node):
        """Removing all the nodes that follow the node"""
        self._successors_all = []
        self._checking_successors_nodes(node=node, remove=False)
        self.remove_nodes_connections(nodes=node)
        nodes_removed = []
        for nd in self._successors_all:
            if nd in self.nodes:
                nodes_removed.append(nd.name)
                self.remove_nodes(nodes=nd, check_ready=False)
                self.remove_previous_connections(nodes=nd)
        return set(nodes_removed)

    def _checking_path(self, node_name, first_name, path=0):
        """Calculate all paths using connections list (re-entering function)."""
        if not self.successors[node_name]:
            return True
        for nd_in in self.successors[node_name]:
            if nd_in.name in self.max_paths[first_name].keys():
                # chose the maximum paths
                self.max_paths[first_name][nd_in.name] = max(
                    self.max_paths[first_name][nd_in.name], path + 1
                )
            else:
                self.max_paths[first_name][nd_in.name] = path + 1
            self._checking_path(
                node_name=nd_in.name, first_name=first_name, path=path + 1
            )

    def calculate_max_paths(self):
        """
        Calculate maximum paths.

        Maximum paths are calculated between any node without "history" (no predecessors)
        and all of the connections.

        """
        self.max_paths = {}
        first_nodes = [key for (key, val) in self.predecessors.items() if not val]
        for nm in first_nodes:
            self.max_paths[nm] = {}
            self._checking_path(node_name=nm, first_name=nm)

    def create_dotfile_simple(self, outdir, name="graph"):
        """creates a simple dotfile (no nested structure)"""

        dotstr = "digraph G {\n"
        for nd in self.nodes:
            if is_workflow(getattr(nd, "_task", None)):
                if nd.state:
                    # adding color for wf with a state
                    dotstr += f"{nd.name} [shape=box, color=blue]\n"
                else:
                    dotstr += f"{nd.name} [shape=box]\n"
            else:
                if nd.state:
                    # adding color for nd with a state
                    dotstr += f"{nd.name} [color=blue]\n"
                else:
                    dotstr += f"{nd.name}\n"
        for ed in self.edges:
            # if the tails nd has a final state, than the state is passed
            # to the next node and the arrow is blue
            if ed[0].state and ed[0].state.splitter_rpn_final:
                dotstr += f"{ed[0].name} -> {ed[1].name} [color=blue]\n"
            else:
                dotstr += f"{ed[0].name} -> {ed[1].name}\n"

        dotstr += "}"
        Path(outdir).mkdir(parents=True, exist_ok=True)
        dotfile = Path(outdir) / f"{name}.dot"
        dotfile.write_text(dotstr)
        return dotfile

    def create_dotfile_detailed(self, outdir, name="graph_det"):
        """creates a detailed dotfile (detailed connections - input/output fields,
        but no nested structure)
        """
        dotstr = "digraph structs {\n"
        dotstr += "node [shape=record];\n"
        if not self._nodes_details:
            raise Exception("node_details is empty, detailed dotfile can't be created")
        for nd_nm, nd_det in self.nodes_details.items():
            if nd_nm == INPUTS_NODE_NAME:  # the main workflow itself
                # wf inputs
                wf_inputs_str = f'{{<{nd_det["outputs"][0]}> {nd_det["outputs"][0]}'
                for el in nd_det["outputs"][1:]:
                    wf_inputs_str += f" | <{el}> {el}"
                wf_inputs_str += "}"
                dotstr += (
                    f"struct_{self.name} [color=red, "
                    f'label="{{WORKFLOW INPUT: | {wf_inputs_str}}}"];\n'
                )
            elif nd_nm == OUTPUTS_NODE_NAME:
                # wf outputs
                wf_outputs_str = f'{{<{nd_det["inputs"][0]}> {nd_det["inputs"][0]}'
                for el in nd_det["inputs"][1:]:
                    wf_outputs_str += f" | <{el}> {el}"
                wf_outputs_str += "}"
                dotstr += (
                    f"struct_{self.name}_out "
                    f'[color=red, label="{{WORKFLOW OUTPUT: | {wf_outputs_str}}}"];\n'
                )
                # connections to the wf outputs
                for con in nd_det["connections"]:
                    dotstr += f"struct_{con[1]}:{con[2]} -> struct_{self.name}_out:{con[0]};\n"
            else:  # elements of the main workflow
                inputs_str = "{INPUT:"
                for inp in nd_det["inputs"]:
                    inputs_str += f" | <{inp}> {inp}"
                inputs_str += "}"
                outputs_str = "{OUTPUT:"
                for out in nd_det["outputs"]:
                    outputs_str += f" | <{out}> {out}"
                outputs_str += "}"
                dotstr += (
                    f"struct_{nd_nm} "
                    f'[shape=record, label="{inputs_str} | {nd_nm} | {outputs_str}"];\n'
                )
                # connections between elements
                for con in nd_det["connections"]:
                    in_conn = self.name if con[1] == INPUTS_NODE_NAME else con[1]
                    out_conn = self.name if con[0] == OUTPUTS_NODE_NAME else con[0]
                    dotstr += (
                        f"struct_{in_conn}:{con[2]} -> struct_{nd_nm}:{out_conn};\n"
                    )
        dotstr += "}"
        Path(outdir).mkdir(parents=True, exist_ok=True)
        dotfile = Path(outdir) / f"{name}.dot"
        dotfile.write_text(dotstr)
        return dotfile

    def create_dotfile_nested(self, outdir, name="graph"):
        """dotfile that includes the nested structures for workflows"""
        dotstr = "digraph G {\ncompound=true \n"
        dotstr += self._create_dotfile_single_graph(nodes=self.nodes, edges=self.edges)
        dotstr += "}"
        Path(outdir).mkdir(parents=True, exist_ok=True)
        dotfile = Path(outdir) / f"{name}.dot"
        dotfile.write_text(dotstr)
        return dotfile

    def _create_dotfile_single_graph(self, nodes, edges):

        wf_asnd = {}
        dotstr = ""
        for nd in nodes:
            if is_workflow(getattr(nd, "_task", None)):
                nd_graph = nd._task.construct().graph()
                wf_asnd[nd.name] = nd_graph
                # for job in nd_graph.nodes:
                #     nd.create_connections(job)
                dotstr += f"subgraph cluster_{nd.name} {{\n" f"label = {nd.name} \n"
                dotstr += self._create_dotfile_single_graph(
                    nodes=nd_graph.nodes, edges=nd_graph.edges
                )
                if nd.state:
                    dotstr += "color=blue\n"
                dotstr += "}\n"
            else:
                if nd.state:
                    dotstr += f"{nd.name} [color=blue]\n"
                else:
                    dotstr += f"{nd.name}\n"

        dotstr_edg = ""
        for ed in edges:
            if ed[0].name in wf_asnd and ed[1].name in wf_asnd:
                head_nd = list(ed[1].nodes)[0].name
                tail_nd = list(ed[0].nodes)[-1].name
                dotstr_edg += (
                    f"{tail_nd} -> {head_nd} "
                    f"[ltail=cluster_{ed[0].name}, "
                    f"lhead=cluster_{ed[1].name}]\n"
                )
            elif ed[0].name in wf_asnd:
                nd_nodes = wf_asnd[ed[0].name].nodes
                tail_nd = list(nd_nodes)[-1].name
                dotstr_edg += (
                    f"{tail_nd} -> {ed[1].name} [ltail=cluster_{ed[0].name}]\n"
                )
            elif ed[1].name in wf_asnd:
                nd_nodes = wf_asnd[ed[1].name].nodes
                head_nd = list(nd_nodes)[0].name
                dotstr_edg += (
                    f"{ed[0].name} -> {head_nd} [lhead=cluster_{ed[1].name}]\n"
                )
            else:
                dotstr_edg += f"{ed[0].name} -> {ed[1].name}\n"
            # if tail has a state and a final splitter (not combined) than the state
            # is passed to the next state and the arrow is blue
            if ed[0].state and ed[0].state.splitter_rpn_final:
                dotstr_edg = dotstr_edg[:-1] + " [color=blue]\n"
        dotstr = dotstr + dotstr_edg
        return dotstr

    def export_graph(self, dotfile, ext="png"):
        """exporting dotfile to other formats requires the dot command"""
        available_ext = [
            "bmp",
            "canon",
            "cgimage",
            "cmap",
            "cmapx",
            "cmapx_np",
            "dot",
            "dot_json",
            "eps",
            "exr",
            "fig",
            "gif",
            "gv",
            "icns",
            "ico",
            "imap",
            "imap_np",
            "ismap",
            "jp2",
            "jpe",
            "jpeg",
            "jpg",
            "json",
            "json0",
            "mp",
            "pct",
            "pdf",
            "pic",
            "pict",
            "plain",
            "plain-ext",
            "png",
            "pov",
            "ps",
            "ps2",
            "psd",
            "sgi",
            "svg",
            "svgz",
            "tga",
            "tif",
            "tiff",
            "tk",
            "vml",
            "vmlz",
            "xdot",
            "xdot1.2",
            "xdot1.4",
            "xdot_json",
        ]
        if ext not in available_ext:
            raise Exception(f"invalid extension - {ext}, chose from {available_ext}")

        dot_check = sp.run(["which", "dot"], stdout=sp.PIPE, stderr=sp.PIPE)
        if not dot_check.stdout:
            raise Exception(f"dot command not available, can't create a {ext} file")

        formatted_dot = dotfile.with_suffix(f".{ext}")
        cmd = f"dot -T{ext} -o {formatted_dot} {dotfile}"
        sp.run(cmd.split(), stdout=sp.PIPE, stderr=sp.PIPE)
        return formatted_dot
